"""
Tempo Markings: note=bpm, text (note=bpm), note=note, (note=note), (note=bpm)
"""

from pathlib import Path

from musicscore import Score, Chord
from musicscore.metronome import Metronome
from unittest import TestCase


class TestLily31c(TestCase):
    def test_lily_31c_MetronomeMarks(self):
        score = Score()
        part = score.add_part("p1")
        [part.add_chord(Chord(72, 2)) for _ in range(5)]
        chords = part.get_chords()

        chords[0].metronome = Metronome(per_minute=100, beat_unit=1.5)
        # chords[1].metronome = Metronome(per_minute=100, beat_unit=8, parenthesis=True, relative_x=40)
        # chords[1].add_words('Adagio', placement='above')
        # chords[2].metronome = Metronome(per_minute=None, beat_unit=1.5, second_beat_unit=3)
        # chords[3].metronome = Metronome(per_minute=None, beat_unit=1.5, second_beat_unit=3, parenthesis=True)
        # chords[0].metronome = Metronome(per_minute=77, beat_unit=1.5, parenthesis=True)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, Chord, Midi
from unittest import TestCase


class TestAccidentals(TestCase):
    def test_modes_in_quarter_steps(self):
        """
        All pitches from c4 to c5 in ascending quarter steps; different Accidental.mode: standard, enharmonic, sharp,
        flat, force-sharp, force-flat.
        """

        def create_part():
            scale = [Midi(m / 2) for m in range(120, 144)]
            p = score.add_part(f"{mode}")
            p.name = f"{mode}"
            for m in scale:
                m.accidental.mode = mode
                ch = Chord(m, 1)
                p.add_chord(ch)

        score = Score("Accidental modes")
        for mode in [
            "standard",
            "enharmonic",
            "sharp",
            "flat",
            "force-sharp",
            "force-flat",
        ]:
            create_part()
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, G, A, B, C, D, E, F, Chord, Midi
from unittest import TestCase

"""
All pitches from G to c”” in ascending steps; First without accidentals, then with a sharp and then with a flat 
accidental. Double alterations and cautionary accidentals are tested at the end.
"""


class TestLily01a(TestCase):
    def test_lily_01a_Pitches_Pitches(self):
        score = Score("Pitches and accidentals")
        p = score.add_part("pitches")
        scale_1 = [
            G(2),
            A(2),
            B(2),
            C(3),
            D(3),
            E(3),
            F(3),
            G(3),
            A(3),
            B(3),
            C(4),
            D(4),
            E(4),
            F(4),
            G(4),
            A(4),
            B(4),
            C(5),
            D(5),
            E(5),
            F(5),
            G(5),
            A(5),
            B(5),
            C(6),
            D(6),
            E(6),
            F(6),
            G(6),
            A(6),
            B(6),
            C(7),
        ]
        # with sharps
        scale_2 = [midi.__deepcopy__() for midi in scale_1]
        for midi in scale_2:
            midi.transpose(1)
            if midi.value % 12 in [0, 5]:
                # B# and E#
                midi.accidental.mode = "force-sharp"
            else:
                midi.accidental.mode = "sharp"
        # with flats
        scale_3 = [midi.__deepcopy__() for midi in scale_1]
        for midi in scale_3:
            midi.transpose(-1)
            if midi.value % 12 in [11, 4]:
                # C-flat and F-flat
                midi.accidental.mode = "force-flat"
            else:
                midi.accidental.mode = "flat"
        # Double alterations and cautionary accidentals.
        cs = [Midi(74), Midi(70), Midi(73), Midi(73), Midi(73), Midi(73)]
        cs[0].accidental.mode = "force-sharp"
        cs[1].accidental.mode = "force-flat"
        for midi in scale_1 + scale_2 + scale_3 + cs:
            p.add_chord(Chord(midis=midi, quarter_duration=1))
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Nested repeats, each with alternative endings.
"""

from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase


class TestLily45d(TestCase):
    def test_lily_45d_Repeats_Nested_Alternatives(self):
        score = Score()
        p = score.add_part("p1")
        [p.add_chord(Chord(0, 4)) for _ in range(11)]

        score.set_multi_measure_rest(3, 5)
        score.set_multi_measure_rest(6, 8)

        p.get_measure(2).set_repeat_barline()
        p.get_measure(2).set_repeat_ending(number=1, type="start")
        p.get_measure(2).set_repeat_ending(number=1, type="stop")

        p.get_measure(3).set_repeat_ending(number=2, type="start")
        p.get_measure(5).set_repeat_ending(number=2, type="discontinue")

        p.get_measure(9).set_repeat_barline()
        p.get_measure(9).set_repeat_ending(number=1, type="start")
        p.get_measure(9).set_repeat_ending(number=1, type="stop")

        p.get_measure(10).set_repeat_ending(number=2, type="start")
        p.get_measure(10).set_repeat_ending(number=5, type="discontinue")

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from typing import List

from musicscore.exceptions import MusicTreeTypeError
from musicscore.util import isinstance_as_string
from verysimpletree.tree import Tree

__all__ = ["MusicTree"]


class MusicTree(Tree):
    """
    MusicTree is the parent class of all music tree objects:
        - :obj:`~musicscore.score.Score` (root)
        - :obj:`~musicscore.part.Part` (1st layer)
        - :obj:`~musicscore.measure.Measure` (2nd layer)
        - :obj:`~musicscore.staff.Staff` (3rd layer)
        - :obj:`~musicscore.voice.Voice` (4th layer)
        - :obj:`~musicscore.beat.Beat` (5th layer)
        - :obj:`~musicscore.chord.Chord`, :obj:`~musicscore.chord.Rest` or :obj:`~musicscore.chord.GraceChord` (6th layer)
        - :obj:`~musicscore.note.Note` (7th layer)
        - :obj:`~musicscore.midi.Midi` (8th layer)
          Midi can represent a pitch or a rest (value=0) and controls accidental sign of the pitch if necessary.
        - :obj:`~musicscore.accidental.Accidental` (9th layer)
    """

    _ATTRIBUTES = {"show_accidental_signs"}

    default_show_accidental_signs = (
        "modern"  #: Class attribute of :obj:`~musicscore.musictree.MusicTree`
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._show_accidental_signs = None

    @staticmethod
    def _check_args_kwargs(args, kwargs, class_name, get_class_name=None):
        for x in args:
            if not isinstance(x, int) or x < 1:
                raise TypeError(f"args {args} must be positive integers")
        for x in kwargs.values():
            if not isinstance(x, int) or x < 1:
                raise TypeError(f"kwargs values {kwargs} must be positive integers")

        def _get_default_keys():
            default_keys = [
                "part_number",
                "measure_number",
                "staff_number",
                "voice_number",
                "beat_number",
                "chord_number",
            ]
            class_names = [
                "Score",
                "Part",
                "Measure",
                "Staff",
                "Voice",
                "Beat",
                "Chord",
            ]
            class_index = class_names.index(class_name)
            get_class_index = (
                -1 if not get_class_name else class_names.index(get_class_name)
            )
            return default_keys[class_index:get_class_index]

        default_keys = _get_default_keys()
        if args and kwargs:
            raise ValueError("Both args and kwargs cannot be set")
        if args:
            if len(args) != len(default_keys):
                raise ValueError(
                    f"Wrong number of args {args}. Keys are: {default_keys}"
                )
            kwargs = {key: value for key, value in zip(default_keys, args)}
        else:
            keys = kwargs.keys()
            if set(keys) != set(default_keys[: len(keys)]):
                raise ValueError(
                    f"Wrong (number) of keys: {list(keys)} in kwargs. All Keys: {default_keys} must be set."
                )
        return kwargs

    def _check_child_to_be_added(self, child):
        if not isinstance_as_string(child, "MusicTree"):
            raise MusicTreeTypeError(
                f"MusicTree child must be of type MusicTree not {child.__class__}"
            )

        parent_child = {
            "Score": "Part",
            "Part": "Measure",
            "Measure": "Staff",
            "Staff": "Voice",
            "Voice": "Beat",
            "Beat": "Chord",
            "GraceChord": "Note",
            "Rest": "Note",
            "Chord": "Note",
            "Note": "Midi",
            "Midi": "Accidental",
            "C": "Accidental",
            "D": "Accidental",
            "E": "Accidental",
            "F": "Accidental",
            "G": "Accidental",
            "A": "Accidental",
            "B": "Accidental",
        }

        try:
            if not isinstance_as_string(child, parent_child[self.__class__.__name__]):
                raise MusicTreeTypeError(
                    f"{self.__class__.__name__} accepts only children of type {parent_child[self.__class__.__name__]} not "
                    f"{child.__class__.__name__}"
                )
        except KeyError:
            raise NotImplementedError(
                f"{self.__class__.__name__} add_child() not implemented."
            )

    def _get_kwargs(self, args_, kwargs_, get_class_name):
        if isinstance_as_string(self, "Score"):
            return self._check_args_kwargs(args_, kwargs_, "Score", get_class_name)
        elif isinstance_as_string(self, "Part"):
            return self._check_args_kwargs(args_, kwargs_, "Part", get_class_name)
        elif isinstance_as_string(self, "Measure"):
            return self._check_args_kwargs(args_, kwargs_, "Measure", get_class_name)
        elif isinstance_as_string(self, "Staff"):
            return self._check_args_kwargs(args_, kwargs_, "Staff", get_class_name)
        elif isinstance_as_string(self, "Voice"):
            return self._check_args_kwargs(args_, kwargs_, "Voice", get_class_name)
        elif isinstance_as_string(self, "Beat"):
            return self._check_args_kwargs(args_, kwargs_, "Beat", get_class_name)
        elif isinstance_as_string(self, "Chord"):
            return self._check_args_kwargs(args_, kwargs_, "Chord", get_class_name)
        else:
            raise MusicTreeTypeError(
                f"MusicTree descendents of type {self.__class__} cannot use this method."
            )

    def _get_music_tree_descendent(self, args, kwargs, get_class_name):
        kwargs = self._get_kwargs(args, kwargs, get_class_name)

        if not kwargs:
            raise TypeError

        if len(kwargs) == 1:
            try:
                return self.get_children()[list(kwargs.values())[0] - 1]
            except IndexError:
                return None
        else:
            output = self
            for key in kwargs:
                string_to_eval = f"output.get_{key.split('_')[0]}(kwargs['{key}'])"
                output = eval(string_to_eval)
                if not output:
                    return None
            return output

    @property
    def show_accidental_signs(self) -> str:
        """
        :obj:`~musicscore.musictree.MusicTree` property

        - If show_accidental_signs is set to None the first get_quantized of ancestors which is ``False`` or ``True`` will be returned.
        - If :obj:`~musicscore.score.Score.show_accidental_signs` is set to None it will be converted to ``default_show_accidental_signs``
        - Possible show_accidental_signs are: None, 'modern', 'traditional'

        :type: Optional[str]
        :rtype: str
        """
        if self._show_accidental_signs is None:
            if self.up:
                return self.up.show_accidental_signs
            else:
                return self.default_show_accidental_signs
        return self._show_accidental_signs

    @show_accidental_signs.setter
    def show_accidental_signs(self, val):
        permitted = [None, "modern", "traditional"]
        if val not in permitted:
            raise ValueError(
                f"show_accidental_signs {val} not in permitted: {permitted}"
            )

        self._show_accidental_signs = val

    def get_beat(self, *args, **kwargs) -> "Beat":
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score` and :obj:`~musicscore.part.Part`, :obj:`~musicscore.measure.Measure` and
        :obj:`~musicscore.staff.Staff` and :obj:`~musicscore.voice.Voice`

        :param args: can be used instead of ``kwargs``. A mixture of args and kwargs is not allowed.
        :param kwargs: ``part_number``, ``measure_number``, ``staff_number``, ``voice_number``, ``beat_number`` depending on
                       musicscore's class. A :obj:`~musicscore.staff.Staff` for example needs ``voice_number`` and ``beat_number`` while a
                       :obj:`~musicscore.score.Score` needs all keyword arguments.
        :rtype: :obj:`~musicscore.beat.Beat`
        """
        return self._get_music_tree_descendent(args, kwargs, "Beat")

    def get_chord(self, *args, **kwargs) -> "Chord":
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score` and :obj:`~musicscore.part.Part`, :obj:`~musicscore.measure.Measure`,
        :obj:`~musicscore.staff.Staff`, :obj:`~musicscore.voice.Voice` and :obj:`~musicscore.beat.Beat`

        :param args: can be used instead of ``kwargs``. A mixture of args and kwargs is not allowed.
        :param kwargs: ``part_number``, ``measure_number``, ``staff_number``, ``voice_number``, ``beat_number``, ``chord_number`` depending on
                       musicscore's class. A :obj:`~musicscore.staff.Staff` for example needs ``voice_number``, ``beat_number`` and
                       ``chord_number`` while a :obj:`~musicscore.score.Score` needs all keyword arguments.
        :rtype: :obj:`~musicscore.chord.Chord`
        """

        return self._get_music_tree_descendent(args, kwargs, "Chord")

    def get_beats(self) -> List["Beat"]:
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score` and :obj:`~musicscore.part.Part`, :obj:`~musicscore.measure.Measure`,
        :obj:`~musicscore.staff.Staff` and :obj:`~musicscore.voice.Voice`.

        :return: a flat list of all beats.
        :rtype: List[:obj:`~musicscore.beat.Beat`]
        """
        if isinstance_as_string(self, "Voice"):
            return self.get_children()
        else:
            output = [ch for child in self.get_children() for ch in child.get_beats()]
            if not output:
                for cls_name in ["Beat", "Chord", "Note", "Midi", "Accidental"]:
                    if isinstance_as_string(self, cls_name):
                        raise MusicTreeTypeError(
                            f"MusicTree descendents of type {self.__class__} cannot use this method."
                        )
            return output

    def get_chords(self) -> List["Chord"]:
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score` and :obj:`~musicscore.part.Part`, :obj:`~musicscore.measure.Measure` and
        :obj:`~musicscore.staff.Staff`, :obj:`~musicscore.voice.Voice` and :obj:`~musicscore.beat.Beat`

        :return: a flat list of all chords.
        :rtype: List[:obj:`~musicscore.chord.Chord`]
        """
        if isinstance_as_string(self, "Beat"):
            return self.get_children()
        else:
            output = [ch for child in self.get_children() for ch in child.get_chords()]
            if not output:
                for cls_name in ["Chord", "Note", "Midi", "Accidental"]:
                    if isinstance_as_string(self, cls_name):
                        raise MusicTreeTypeError(
                            f"MusicTree descendents of type {self.__class__} cannot use this method."
                        )
            return output

    def get_measure(self, *args, **kwargs) -> "Measure":
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score` and :obj:`~musicscore.part.Part`

        :param args: can be used instead of ``kwargs``. A mixture of args and kwargs is not allowed.
        :param kwargs: ``part_number``, ``measure_number`` depending on musicscore's class.
        :rtype: :obj:`~musicscore.measure.Measure`
        """

        return self._get_music_tree_descendent(args, kwargs, "Measure")

    def get_part(self, *args, **kwargs) -> "Part":
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score`

        :param args: can be used instead of ``kwargs``. A mixture of args and kwargs is not allowed.
        :param kwargs: ``part_number``.
        :rtype: :obj:`~musicscore.part.Part`
        """

        return self._get_music_tree_descendent(args, kwargs, "Part")

    def get_staff(self, *args, **kwargs) -> "Staff":
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score`, :obj:`~musicscore.part.Part` and :obj:`~musicscore.measure.Measure`

        :param args: can be used instead of ``kwargs``. A mixture of args and kwargs is not allowed.
        :param kwargs: ``part_number``, ``measure_number``, ``staff_number`` depending on musicscore's class.
        :rtype: :obj:`~musicscore.staff.Staff`
        """

        return self._get_music_tree_descendent(args, kwargs, "Staff")

    def get_voice(self, *args, **kwargs) -> "Voice":
        """
        :obj:`~musicscore.musictree.MusicTree` method

        This method can be used for :obj:`~musicscore.score.Score` and :obj:`~musicscore.part.Part`, :obj:`~musicscore.measure.Measure` and
        :obj:`~musicscore.staff.Staff`

        :param args: can be used instead of ``kwargs``. A mixture of args and kwargs is not allowed.
        :param kwargs: ``part_number``, ``measure_number``, ``staff_number``, ``voice_number`` depending on musicscore's class.
        :rtype: :obj:`~musicscore.voice.Voice`
        """
        return self._get_music_tree_descendent(args, kwargs, "Voice")



from musicscore import QuarterDuration, Part, Chord
from musicscore.exceptions import MetronomeWrongBeatUnitError
from musicscore.metronome import Metronome

# from unittest import TestCase
from musicxml.xmlelement.xmlelement import XMLSound
from unittest import TestCase


class TestCase(TestCase):
    def test_metronome_init(self):
        m = Metronome(100)
        assert isinstance(m.beat_unit, QuarterDuration)
        assert m.per_minute == 100
        assert m.xml_per_minute.value_ == "100"
        assert m.beat_unit == 1
        assert m.xml_beat_unit.value_ == "quarter"
        assert len(m.get_xml_beat_dot_objects()) == 0

    def test_metronome_dotted(self):
        m = Metronome(100, beat_unit=1.5)
        assert m.xml_beat_unit.value_ == "quarter"
        assert len(m.get_xml_beat_dot_objects()) == 1

        m = Metronome(100, beat_unit=2 + 1 + 0.5)
        assert m.xml_beat_unit.value_ == "half"
        assert len(m.get_xml_beat_dot_objects()) == 2

    def test_metronome_change_values(self):
        m = Metronome(100)
        assert m.xml_beat_unit.value_ == "quarter"
        assert len(m.get_xml_beat_dot_objects()) == 0
        m.beat_unit = 3
        assert m.xml_beat_unit.value_ == "half"
        assert len(m.get_xml_beat_dot_objects()) == 1

    def test_metronome_wrong_beat_units(self):
        with self.assertRaises(MetronomeWrongBeatUnitError):
            Metronome(100, 1.1)
        with self.assertRaises(MetronomeWrongBeatUnitError):
            Metronome(100, 2 / 3)
        with self.assertRaises(MetronomeWrongBeatUnitError):
            Metronome(100, 6 / 5)

    def test_metronome_sound(self):
        m = Metronome(100)
        assert isinstance(m.sound, XMLSound)
        assert m.sound.tempo == 100
        m = Metronome(100, beat_unit=1.5)
        assert m.sound.tempo == 150

    def test_add_metronome_to_measure(self):
        p = Part("p1")
        ch = Chord(60, 4)
        p.add_chord(ch)
        ch.metronome = Metronome(100, 1.5)
        expected = """<direction placement="above">
      <direction-type>
        <metronome>
          <beat-unit>quarter</beat-unit>
          <beat-unit-dot />
          <per-minute>100</per-minute>
        </metronome>
      </direction-type>
      <sound tempo="150" />
    </direction>
"""
        p.finalize()
        assert p.get_measure(1).xml_direction.to_string() == expected


"""
      <direction placement="above">
        <direction-type>
          <words font-size="12" font-weight="normal" relative-y="30">Andante </words>
        </direction-type>
        <direction-type>
          <metronome font-family="EngraverTextT" font-size="12" parentheses="yes" relative-y="30" relative-x="90">
            <beat-unit>quarter</beat-unit>
            <per-minute font-family="Times New Roman" font-size="12">80</per-minute>
          </metronome>
        </direction-type>
        <sound tempo="80"/>
      </direction>
"""

"""
      <direction placement="above">
        <direction-type>
          <metronome font-family="Finale Maestro" font-size="12" parentheses="yes" relative-y="30">
            <beat-unit>quarter</beat-unit>
            <beat-unit-dot/>
            <beat-unit>quarter</beat-unit>
          </metronome>
        </direction-type>
      </direction>
"""

"""
      <direction placement="above">
        <direction-type>
          <metronome font-family="Finale Maestro" font-size="12" parentheses="yes" relative-y="30">
            <beat-unit>quarter</beat-unit>
            <per-minute font-family="Times New Roman" font-size="12">120</per-minute>
          </metronome>
        </direction-type>
        <sound tempo="120"/>
      </direction>
"""

"""
      <direction placement="above">
        <direction-type>
          <metronome font-family="Finale Maestro" font-size="12" relative-y="30">
            <beat-unit>quarter</beat-unit>
            <beat-unit>half</beat-unit>
          </metronome>
        </direction-type>
      </direction>
"""



import inspect
from itertools import permutations

from musicscore import Score, Time, C, Chord
from unittest import TestCase

from musicscore.tests.util import generate_path

patterns = [
    (1, 1),
    (1, 3),
    (1, 1, 2),
    (1, 7),
    (1, 1, 6),
    (1, 2, 5),
    (1, 3, 4),
    (1, 1, 1, 5),
    (1, 1, 2, 4),
    (1, 1, 3, 3),
    (1, 2, 2, 3),
    (1, 1, 1, 1, 4),
    (1, 1, 1, 2, 3),
    (1, 2, 2, 2, 1),
    (1, 1, 1, 1, 1, 3),
    (1, 1, 1, 1, 2, 2),
    (1, 1, 1, 1, 1, 1, 2),
]
permuted_patterns = [list(set(list(permutations(pattern)))) for pattern in patterns]


class TestBeams(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part("p1")

    def test_beams_complicated_rhythm(self):
        self.part.add_measure(Time(1, 4))
        for list_of_patterns in permuted_patterns:
            for pattern in list_of_patterns:
                for index, qd in enumerate([x / sum(pattern) for x in pattern]):
                    ch = Chord(C(5), qd)
                    if index == 0:
                        ch.add_lyric(pattern, justify="left")
                    self.part.add_chord(ch)

        path = generate_path(inspect.currentframe())
        self.score.export_xml(path)

    # def test_beams_arbitrary_groups(self):
    #     qds = None



from pathlib import Path

from musicscore import Score, Time, Chord, B
from unittest import TestCase

"""
Compound time signatures with separate fractions displayed: 3/8+2/8+3/4 and 5/2+1/8.
"""


class TestLily11d(TestCase):
    def test_lily_11d_TimeSignatures_CompoundMultiple(self):
        score = Score()
        part = score.add_part("part")
        part.add_measure(time=Time(3, 8, 2, 8, 3, 4))
        part.add_measure(time=Time(5, 2, 1, 8))
        [part.add_chord(Chord(B(4), x / 2)) for x in [1, 1, 1, 1, 1, 2, 2, 2]]
        [part.add_chord(Chord(B(4), x)) for x in [8, 2, 0.5]]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from typing import List

from math import trunc
from fractions import Fraction

from musicscore.chord import _split_copy, _group_chords, Chord
from musicscore.config import SPLITTABLES, GENERALSPLITTABLES, SPLITTEXCEPTIONS
from musicscore.exceptions import (
    BeatWrongDurationError,
    BeatIsFullError,
    BeatHasNoParentError,
    ChordHasNoQuarterDurationError,
    ChordHasNoMidisError,
    AlreadyFinalizedError,
    BeatNotFullError,
    AddChordError,
    QuarterDurationIsNotWritable,
    BeatUpdateChordTupletsError,
    ChordTypeNotSetError,
)
from musicscore.finalize import FinalizeMixin
from musicscore.musictree import MusicTree
from musicscore.quantize import QuantizeMixin
from musicscore.quarterduration import QuarterDuration, QuarterDurationMixin
from musicscore.tuplet import Tuplet
from musicscore.util import lcm, split_list

__all__ = ["Beat", "beam_chord_group", "get_chord_group_subdivision"]


def _convert_to_quarter_duration_splittables_dictionary(simple_splittalbes):
    output = {}
    for key, value in simple_splittalbes.items():
        output[QuarterDuration(*key)] = {
            QuarterDuration(*k): [QuarterDuration(*qd) for qd in v]
            for k, v in value.items()
        }
    return output


_SPLITTABLE_QUARTER_DURATIONS = _convert_to_quarter_duration_splittables_dictionary(
    SPLITTABLES
)


def _find_nearest_quantized_value(quantized_values, values):
    output = []
    for value in values:
        nearest_quantized = min(
            enumerate(quantized_values), key=lambda x: abs(x[1] - value)
        )[1]
        delta = nearest_quantized - value
        output.append((nearest_quantized, delta))
    return output


def _find_q_delta(quantized_locations, values):
    qs = _find_nearest_quantized_value(quantized_locations, values)
    d = 0
    for q in qs:
        d += abs(q[1])
    return d


def _find_quantized_locations(duration, subdivision):
    output = range(subdivision + 1)
    fr = duration / subdivision
    output = [x * fr for x in output]
    return output


def get_chord_group_subdivision(chords):
    qds = [ch.quarter_duration for ch in chords if ch.quarter_duration != 0]
    if len(qds) == 1:
        return qds[0].denominator
    qd_sum = sum(qds)
    if qd_sum in [3 / 2, 3 / 4, 3, 6]:
        denominators = list(dict.fromkeys([qd.denominator for qd in qds]))
        l_c_m = lcm(denominators)
        if l_c_m not in [1, 2, 4, 8, 16]:
            raise NotImplementedError
        else:
            return l_c_m
    permitted_sums = [1 / 4, 1 / 2, 1, 2, 4, 8]
    if qd_sum not in permitted_sums:
        raise ValueError(f"sum of chords {qd_sum} must be in {permitted_sums}")
    qds = [qd / qd_sum for qd in qds]
    denominators = list(dict.fromkeys([qd.denominator for qd in qds]))
    if len(denominators) > 1:
        l_c_m = lcm(denominators)
        if l_c_m not in denominators and l_c_m > 16:
            return None
        else:
            return l_c_m
    else:
        return denominators[0]


def beam_chord_group(chord_group: List["Chord"]) -> None:
    # print('setting beams', [ch.quarter_duration for ch in chord_group])
    """
    Function for setting beams of a list of chords (chord_group). This function is used to create or update beams inside a beat.
    Chord types must be set first.

    """
    chord_group = [ch for ch in chord_group if ch.quarter_duration != 0]
    for ch in chord_group:
        if ch.type is None:
            raise ChordTypeNotSetError(
                "Beaming chord groups not possible if chord types are not set."
            )

    def remove_rests_from_both_ends(chords):
        is_rest_list = [ch.is_rest for ch in chords]
        if False not in is_rest_list:
            return []
        if True in is_rest_list:
            first_non_chord_index = is_rest_list.index(False)
            output = chords[first_non_chord_index:]
            is_rest_list = is_rest_list[first_non_chord_index:]
            if is_rest_list[-1] is True:
                last_non_chord_index = next(
                    i
                    for i in reversed((range(len(is_rest_list))))
                    if is_rest_list[i] is True
                )
                output = output[:last_non_chord_index]
            return output
        else:
            return chords

    chord_group = remove_rests_from_both_ends(chord_group)
    if not chord_group:
        return

    # valid values for XMLBeam are: 'begin', 'continue', 'end', 'forward hook', 'backward hook'
    def add_beam_to_chord(chord, number, value):
        if chord.beams is not None and not chord.beams.get(number):
            chord.set_beam(number, value)

    def add_last_beam(chord, last_number_of_beams, current_number_of_beams, cont=False):
        """
        cont=True will be for example used to continue one beam (eight beam) in the middle of the beat with 32nds.
        """
        if last_number_of_beams <= current_number_of_beams:
            if cont:
                add_beam_to_chord(chord, 1, "continue")
                for num in range(2, last_number_of_beams + 1):
                    add_beam_to_chord(chord, num, "end")
            else:
                for n in range(1, last_number_of_beams + 1):
                    add_beam_to_chord(chord, n, "end")
        else:
            if current_number_of_beams != 0:
                if cont:
                    add_beam_to_chord(chord, 1, "continue")
                    for n in range(2, current_number_of_beams + 1):
                        add_beam_to_chord(chord, n, "end")
                else:
                    for n in range(1, current_number_of_beams + 1):
                        add_beam_to_chord(chord, n, "end")
                for n in range(current_number_of_beams + 1, last_number_of_beams + 1):
                    add_beam_to_chord(chord, n, "backward")

    current_number_of_beams = 0

    # adding all necessary beams to all notes save the notes of the last chord. For last chord add_last_beam() will be used.
    index = 0
    while index != len(chord_group) - 1:
        # print('index', index)
        # print('current', current_number_of_beams)
        chord = chord_group[index]
        next_chord = chord_group[index + 1]
        # types is a list of tuples with three values: (beam value, number of the begining beam, number of the ending beam)
        types = []
        b1, b2 = chord.number_of_beams, next_chord.number_of_beams
        # print(chord.quarter_duration, next_chord.quarter_duration)
        # print(b1, b2, current_number_of_beams)
        if chord.is_rest:
            pass
            "do nothing"
        else:
            if not b1:
                pass
                "do nothing"
            elif next_chord.is_rest:
                if current_number_of_beams == 0:
                    types.append(("begin", 1, 1))
                    if b1 > 1:
                        types.append(("forward", 1, b1))
                else:
                    types.append(("continue", 1, 1))
                    if b1 > 1:
                        if b1 <= current_number_of_beams:
                            types.append(("end", 2, b1))
                        else:
                            types.append(("end", 2, current_number_of_beams))
                            types.append(("forward", current_number_of_beams + 1, b1))

                current_number_of_beams = 1
            else:
                "do something regular"
                if b1 and not b2:
                    add_last_beam(chord, b1, current_number_of_beams)
                else:
                    if b2 < b1 <= current_number_of_beams:
                        # print('huhu')
                        types.append(("continue", 1, b2))
                        types.append(("end", b2 + 1, b1))
                    elif b2 < b1 > current_number_of_beams:
                        if current_number_of_beams == 0:
                            types.append(("begin", 1, b2))
                            types.append(("forward", b2 + 1, b1))
                        else:
                            if current_number_of_beams > b2:
                                types.append(("continue", 1, b2))
                                types.append(("end", b2 + 1, current_number_of_beams))
                                types.append(
                                    ("backward", current_number_of_beams + 1, b1)
                                )
                            elif current_number_of_beams < b2:
                                types.append(("continue", 1, current_number_of_beams))
                                types.append(("begin", current_number_of_beams + 1, b2))
                                types.append(("forward", b2 + 1, b1))
                            else:
                                types.append(("continue", 1, b2))
                                types.append(("forward", b2 + 1, b1))
                            #
                            # if (chord.quarter_duration == QuarterDuration(1, 6) and chord.offset == QuarterDuration(1,
                            #                                                                                         3)
                            # ) or (
                            #         chord.quarter_duration == QuarterDuration(1, 8) and chord.offset == QuarterDuration(
                            #     3, 8)
                            # ):
                            #     types.append(('backward', b2 + 1, b1))
                            # else:
                            #     types.append(('forward', b2 + 1, b1))

                        # current_number_of_beams = b1
                    elif b2 == b1 <= current_number_of_beams:
                        types.append(("continue", 1, b1))
                        # current_number_of_beams = b1

                    elif b2 == b1 > current_number_of_beams:
                        if current_number_of_beams == 0:
                            types.append(("begin", 1, b2))
                        else:
                            types.append(("continue", 1, current_number_of_beams))
                            types.append(("begin", current_number_of_beams + 1, b2))

                    elif b2 > b1 <= current_number_of_beams:
                        types.append(("continue", 1, b1))

                    elif b2 > b1 > current_number_of_beams:
                        if current_number_of_beams == 0:
                            types.append(("begin", 1, b1))
                        else:
                            types.append(("continue", 1, current_number_of_beams))
                            types.append(("begin", current_number_of_beams + 1, b1))
        for t in types:
            for num in range(t[1], t[2] + 1):
                add_beam_to_chord(chord, num, t[0])
        if chord.beams:
            current_number_of_beams = len(
                [v for v in chord.beams.values() if v in ["continue", "begin"]]
            )
            # print('setting current to', current_number_of_beams)
        if index == len(chord_group) - 2 and b2:
            add_last_beam(next_chord, b2, current_number_of_beams)
        index += 1


class Beat(MusicTree, QuarterDurationMixin, QuantizeMixin, FinalizeMixin):
    """
    Parent type: :obj:`~musicscore.voice.Voice`

    Child type: :obj:`~musicscore.chord.Chord`

    Beat is the direct ancestor of chords. Each :obj:`~musicscore.chord.Chord` is placed with an offset between 0 and beat's
    quarter duration inside the beat as its child .

    Quarter duration of a beat's :obj:`~musicscore.chord.Chord` child can exceed its own quarter duration. If a
    :obj:`~musicscore.chord.Chord` is longer than the quarter duration of beat's parent :obj:`~musicscore.voice.Voice`,
    a leftover :obj:`~musicscore.chord.Chord` will be added as leftover property to the :obj:`~musicscore.voice.Voice` which will be added
    to next measure's appropriate voice .

    Beat manages splitting of each child :obj:`~musicscore.chord.Chord` into appropriate tied :obj:`~musicscore.chord.Chord` s if needed,
    for example if this chord has a non-writable quarter duration like 5/6.

    The dots and tuplets are also added here to :obj:`~musicscore.chord.Chord` or directly to their :obj:`~musicscore.note.Note` children.

    Beaming and quantization are also further important tasks of a beat.
    """

    _PERMITTED_DURATIONS = {4, 2, 1, 0.5}

    def __init__(self, quarter_duration=1):
        super().__init__(quarter_duration=quarter_duration)
        self._filled_quarter_duration = 0
        self.leftover_chord = None
        self._subdivision = None

    def _add_child(self, child):
        child._parent = self
        self._children.append(child)
        try:
            self.up.up.up.up.set_current_measure(
                staff_number=self.up.up.number,
                voice_number=self.up.number,
                measure=self.up.up.up,
            )
        except AttributeError:
            pass

    def _add_chord(self, chord=None):
        if chord is None:
            chord = Chord(midis=60, quarter_duration=self.quarter_duration)
        return self.add_child(chord)

    def _change_children_quarter_durations(self, quarter_durations):
        if len(quarter_durations) != len(self.get_children()):
            raise ValueError
        if sum(quarter_durations) != self.quarter_duration:
            raise ValueError
        for qd, ch in zip(quarter_durations, self.get_children()):
            ch._quarter_duration = qd

    def _check_permitted_duration(self, val):
        for d in self._PERMITTED_DURATIONS:
            if val == d:
                return
        raise BeatWrongDurationError(f"Beat's quarter duration {val} is not allowed.")

    def _get_quantized_locations(self, subdivision):
        return _find_quantized_locations(self.quarter_duration, subdivision)

    def _get_quantized_quarter_durations(self, quarter_durations):
        if sum(quarter_durations) != self.quarter_duration:
            raise ValueError(
                f"Sum of quarter_durations '{quarter_durations}: {sum(quarter_durations)}' is not equal to beat quarter_duration "
                f"'{self.quarter_duration}'"
            )

        def _get_positions():
            output = [0]
            for i, qd in enumerate(quarter_durations):
                output.append(output[i] + qd)
            return output

        positions = _get_positions()
        permitted_divs = self.get_possible_subdivisions()[:]
        best_div = permitted_divs.pop(0)
        last_q_delta = _find_q_delta(
            self._get_quantized_locations(subdivision=best_div), positions
        )

        for div in permitted_divs:
            current_q_delta = _find_q_delta(
                self._get_quantized_locations(subdivision=div), positions
            )

            if current_q_delta < last_q_delta:
                best_div = div
                last_q_delta = current_q_delta

            elif (current_q_delta == last_q_delta) and (div < best_div):
                best_div = div

        quantized_positions = [
            f[0]
            for f in _find_nearest_quantized_value(
                self._get_quantized_locations(subdivision=best_div), positions
            )
        ]

        quantized_durations = []

        for i in range(len(quarter_durations)):
            fr = Fraction(
                quantized_positions[i + 1] - quantized_positions[i]
            ).limit_denominator(trunc(best_div / self.quarter_duration))
            quantized_durations.append(QuarterDuration(fr))
        return quantized_durations

    @staticmethod
    def _split_chord(chord, quarter_durations):
        output = [chord]
        chord._quarter_duration = quarter_durations[0]
        for qd in quarter_durations[1:]:
            copied = _split_copy(chord, qd)
            output.append(copied)
        for index, ch in enumerate(output[:-1]):
            next_ch = output[index + 1]
            ch.add_tie("start")
            next_ch.add_tie("stop")
            for midi in next_ch.midis:
                midi.accidental.show = False
        return output

    def _split_not_writable(self, chord, offset):
        starting_ties = []
        for midi in chord.midis:
            starting_ties.append(True if midi.is_tied_to_next else False)

        def _get_quarter_durations():
            if _SPLITTABLE_QUARTER_DURATIONS.get(
                offset
            ) and _SPLITTABLE_QUARTER_DURATIONS.get(offset).get(chord.quarter_duration):
                quarter_durations = _SPLITTABLE_QUARTER_DURATIONS.get(offset).get(
                    chord.quarter_duration
                )
                if quarter_durations:
                    return [QuarterDuration(qd) for qd in quarter_durations]

            if GENERALSPLITTABLES.get(chord.quarter_duration.numerator):
                quarter_durations = [
                    QuarterDuration(x, chord.quarter_duration.denominator)
                    for x in GENERALSPLITTABLES.get(chord.quarter_duration.numerator)
                ]
                return quarter_durations

            try:
                split_qds = (
                    SPLITTEXCEPTIONS.get(self.quarter_duration.value)
                    .get(self.get_subdivision())
                    .get(chord.quarter_duration.as_integer_ratio())
                )
                if split_qds:
                    quarter_durations = [
                        QuarterDuration(qd[0], qd[1]) for qd in split_qds
                    ]
                    return quarter_durations
            except AttributeError:
                return None

        quarter_durations = _get_quarter_durations()
        if not quarter_durations:
            return [chord]
        output = self._split_chord(chord, quarter_durations)
        for midi, tied in zip(output[-1].midis, starting_ties):
            if tied is True:
                midi.add_tie("start")
        return output

    def _update_chord_types(self):
        for ch in self.get_chords():
            if not ch.type:
                if self.get_subdivision() and not ch.quarter_duration.beat_subdivision:
                    ch.quarter_duration.beat_subdivision = self.get_subdivision()
                try:
                    ch.type = ch.quarter_duration.get_type()
                except QuarterDurationIsNotWritable as err:
                    raise QuarterDurationIsNotWritable(
                        f"Chord {ch.get_position_in_tree()} with offset {ch.offset}: {err} Consider setting type, number_of_dots and tuplet properties of the chord manually or splitting it into writable chords."
                    )

    def _update_chord_number_of_dots(self):
        for ch in self.get_chords():
            if ch.number_of_dots is None:
                if self.get_subdivision() and not ch.quarter_duration.beat_subdivision:
                    ch.quarter_duration.beat_subdivision = self.get_subdivision()
                ch.number_of_dots = ch.quarter_duration.get_number_of_dots()

    def _update_chord_tuplets(self):
        non_grace_chords = [
            chord for chord in self.get_chords() if chord.quarter_duration != 0
        ]
        if None not in {ch.tuplet for ch in non_grace_chords}:
            return
        if {ch.tuplet for ch in non_grace_chords} != {None}:
            raise BeatUpdateChordTupletsError(
                "Beat cannot manage tuplets automatically if it contains chords with manually set tuplet properties."
            )

        def _update_tuplets(chord_group, actual_notes, quarter_duration=1):
            if actual_notes <= 16 or actual_notes == 32:
                if actual_notes not in [1, 2, 4, 8, 16, 32]:
                    for chord in chord_group:
                        chord.tuplet = Tuplet(
                            actual_notes=actual_notes, quarter_duration=quarter_duration
                        )
                        if chord == chord_group[0]:
                            chord.tuplet.bracket_type = "start"
                        elif chord == chord_group[-1]:
                            chord.tuplet.bracket_type = "stop"
                        else:
                            pass
            else:
                raise NotImplementedError(
                    "tuplets of actual_notes > 16 cannot be implemented."
                )

        actual_notes = get_chord_group_subdivision(non_grace_chords)
        if not actual_notes:
            if self.quarter_duration == 1:
                grouped_chords = _group_chords(self.get_children(), [1 / 2, 1 / 2])
                if grouped_chords:
                    for g in grouped_chords:
                        actual_notes = get_chord_group_subdivision(g)
                        _update_tuplets(g, actual_notes, 1 / 2)
                    return
                else:
                    raise NotImplementedError(
                        "Beat cannot be halved. It cannot manage the necessary grouping of chords."
                    )
            else:
                raise NotImplementedError(
                    "Beat with quarter_duration other than one cannot manage more than one group of chords."
                )

        _update_tuplets(non_grace_chords, actual_notes, self.quarter_duration)

    def _update_chord_beams(self):
        chords = self.get_chords()
        if chords:
            ## update types
            for ch in chords:
                if ch.type is None:
                    self._update_chord_types()
            ## group chords
            chord_groups = None
            continue_eighth_beam = False
            # group inside beat
            if self.get_subdivision() == 8 and self.quarter_duration == 1:
                chord_groups = _group_chords(chords, [1 / 2, 1 / 2])
                if chord_groups and 1 not in {len(group) for group in chord_groups}:
                    continue_eighth_beam = True
                else:
                    chord_groups = None
            if not chord_groups:
                chord_groups = [chords]
            # break beams
            broken_chord_groups = []
            for group in chord_groups:
                split_indices = [
                    index for index, chord in enumerate(group) if chord.broken_beam
                ]
                for broken_group in split_list(group, split_indices):
                    broken_chord_groups.append(broken_group)
            # create beams
            for group in broken_chord_groups:
                beam_chord_group(group)
            # continue eighth beam
            if continue_eighth_beam and 1 not in {
                len(group) for group in broken_chord_groups
            }:
                for index, group in enumerate(broken_chord_groups):
                    if index < len(broken_chord_groups) - 1:
                        group[-1].beams[1] = "continue"
                    if index > 0:
                        group[0].beams[1] = "continue"

    def _remove_zero_quarter_durations(self):
        def _get_next_chord(chord):
            next_chord = chord.next
            if not next_chord:
                next_beat = chord.up.next
                while next_beat and not next_chord:
                    try:
                        next_chord = next_beat.get_children()[0]
                    except IndexError:
                        next_beat = next_beat.next
            if not next_chord:
                voice_number = chord.up.up.number
                staff_number = chord.up.up.up.number
                if not staff_number:
                    staff_number = 1
                next_measure = chord.up.up.up.up.next
                if next_measure:
                    next_measure_voice = next_measure.get_chord(
                        staff_number=staff_number, voice_number=voice_number
                    )
                    if next_measure_voice:
                        next_beat = next_measure_voice.get_children()[0]
                        while next_beat and not next_chord:
                            try:
                                next_chord = next_beat.get_children()[0]
                            except IndexError:
                                next_beat = next_beat.next

            return next_chord

        def _get_previous_chord(chord):
            previous_chord = chord.previous
            if not previous_chord:
                previous_beat = chord.up.previous
                while previous_beat and not previous_chord:
                    try:
                        previous_chord = previous_beat.get_children()[-1]
                    except IndexError:
                        previous_beat = previous_beat.previous
            if not previous_chord:
                voice_number = chord.up.up.number
                staff_number = chord.up.up.up.number
                if not staff_number:
                    staff_number = 1
                previous_measure = chord.up.up.up.up.previous
                if previous_measure:
                    previous_measure_voice = previous_measure.get_chord(
                        staff_number=staff_number, voice_number=voice_number
                    )
                    if previous_measure_voice:
                        previous_beat = previous_measure_voice.get_children()[-1]
                        while previous_beat and not previous_chord:
                            try:
                                previous_chord = previous_beat.get_children()[-1]
                            except IndexError:
                                previous_beat = previous_beat.previous

            return previous_chord

        zeros = [ch for ch in self.get_children() if ch.quarter_duration == 0]
        for ch in zeros:
            if ch.all_midis_are_tied_to_next:
                next_chord = _get_next_chord(ch)
                # next_chord.add_lyric('I am next')
                if next_chord:
                    [m.remove_tie("stop") for m in next_chord.midis]
                    next_chord._xml_direction_types = ch._xml_direction_types
                    next_chord._xml_directions = ch._xml_directions
                    next_chord._xml_lyrics = ch._xml_lyrics
                    next_chord._xml_articulations = ch._xml_articulations
                    next_chord._xml_technicals = ch._xml_technicals
                    next_chord._xml_ornaments = ch._xml_ornaments
                    next_chord._xml_dynamics = ch._xml_dynamics
                    next_chord._xml_other_notations = ch._xml_other_notations
                    next_chord._note_attributes = ch._note_attributes
                ch.up.remove(ch)

            elif ch.all_midis_are_tied_to_previous:
                previous_chord = _get_previous_chord(ch)
                if previous_chord:
                    [m.remove_tie("start") for m in previous_chord.midis]
                ch.up.remove(ch)
            else:
                pass

    def _split_not_writable_chords(self) -> None:
        """
        This method checks if the quarter duration of all children chords must be split according to :obj:`~musicscore.beat.SPLITTABLES`
        dictionary. If chord's offset and its quarter duration exist in the dictionary a list of splitting quarter durations can be
        accessed like this: ``SPLITTABLES[chord.offset[chord.quarter_duration]]`` This dictionary can be manipulated by user during runtime
        if needed. Be careful with not writable quarter durations which have to be split (for example 5/6 must be split to 3/6,
        2/6 or some other writable quarter durations).

        :obj:`~musicscore.measure.Measure.finalize()` loops over all its beats calls this method.
        """
        for chord in self.get_children()[:]:
            split = self._split_not_writable(chord, chord.offset)
            if split:
                for ch in split:
                    ch._parent = self
                if chord == self.get_children()[-1]:
                    self._children = self.get_children()[:-1] + split
                else:
                    index = self.get_children().index(chord)
                    self._children = (
                        self.get_children()[:index]
                        + split
                        + self.get_children()[index + 1 :]
                    )

    @property
    def is_filled(self) -> bool:
        """
        :return: ``True`` if no children can be added anymore. If ``False`` there is still room for further child or children.
        :rtype: bool
        """
        if self.filled_quarter_duration == self.quarter_duration:
            return True
        else:
            return False

    @property
    def filled_quarter_duration(self):
        """
        :return: How much of beat's quarter duration is already filled.
        :rtype: QuarterDuration
        """
        return self._filled_quarter_duration

    @property
    def number(self) -> int:
        """
        :return: Beat's number inside its parent's :obj:`musicscore.voice.Voice`
        :rtype: int
        """
        return self.up.get_children().index(self) + 1

    @property
    def offset(self) -> QuarterDuration:
        """
        :return: Offset in Beat's parent :obj:`musicscore.voice.Voice`
        :rtype: QuarterDuration
        """
        if not self.up:
            return None
        elif self.previous is None:
            return 0
        else:
            return self.previous.offset + self.previous.quarter_duration

    def add_child(self, child: Chord) -> List["Chord"]:
        """
        If child's quarter duration is less than beat's remaining quarter duration: child is added to the beat.

        If child's quarter duration is greater than beat's remaining quarter duration: :obj:`~musicscore.chord.Chord`'s :obj:`~musicscore.chord.Chord._split_and_add_beatwise` is
        method called. It is possible to add a chord with a quarter duration exceeding the beat's quarter duration without splitting the chord.
        For example if the first beat in a 4/4 measure gets a chord with quarter duration 3, the chord will be added to this first beat as a
        child and the following two beats will be set to filled without having a child themselves and the parent
        :obj:`~musicscore.voice.Voice` returns the fourth beat if its :obj:`~musicscore.voice.Voice.get_current_beat` is called.

        If child's quarter duration exceeds the :obj:`~musicscore.voice.Voice`'s remaining quarter duration a leftover :obj:`~musicscore.chord.Chord` will be added to the voice and can be
        accessed when the next :obj:`~musicscore.measure.Measure` is created.

        :param child: :obj:`~musicscore.chord.Chord` to be added as child
        :return: list of split chords
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_child")
        self._check_child_to_be_added(child)
        if not self.up:
            raise BeatHasNoParentError(
                "A child Chord can only be added to a beat if it has a voice parent."
            )
        if child.quarter_duration is None:
            raise ChordHasNoQuarterDurationError(
                "Chord with no quarter_duration cannot be added to Beat."
            )
        if not child.midis:
            raise ChordHasNoMidisError("Chord with no midis cannot be added to Beat.")
        if self.is_filled and child.quarter_duration != 0:
            raise BeatIsFullError()
        diff = child.quarter_duration - (
            self.quarter_duration - self.filled_quarter_duration
        )
        if diff <= 0:
            self._filled_quarter_duration += child.quarter_duration
            self._add_child(child)
            return [child]
        else:
            if child.split:
                remaining_quarter_duration = child.quarter_duration
                current_beat = self
                while remaining_quarter_duration and current_beat:
                    if current_beat.quarter_duration < remaining_quarter_duration:
                        current_beat._filled_quarter_duration += (
                            current_beat.quarter_duration
                        )
                        remaining_quarter_duration -= current_beat.quarter_duration
                        current_beat = current_beat.next
                    else:
                        current_beat._filled_quarter_duration += (
                            remaining_quarter_duration
                        )
                        break
                self._add_child(child)
                return [child]
            else:
                beats = self.up.get_children()[self.up.get_children().index(self) :]
                return child._split_and_add_beatwise(beats)

    def add_chord(self, *args, **kwargs):
        raise AddChordError

    def fill_with_rests(self) -> None:
        """
        If :obj:`~musicscore.beat.Beat` is not filled, it will be filled with rest(s)
        """
        if not self.is_filled:
            self._add_chord(
                Chord(
                    0,
                    self.quarter_duration
                    - sum([ch.quarter_duration for ch in self.get_chords()]),
                )
            )

    def finalize(self):
        """
        finalize can only be called once.

        - It calls finalize method of all :obj:`~musicscore.chord.Chord` children.

        - Following updates are triggered: _update_note_tuplets, _update_chord_beams, quantize_quarter_durations (if get_quantized is
          True), _split_not_writable_chords
        """
        if self._finalized:
            raise AlreadyFinalizedError(self)
        if self.is_filled is False:
            self.fill_with_rests()

        if self.get_children():
            self._update_chord_types()
            self._update_chord_number_of_dots()
            self._update_chord_tuplets()
            self._update_chord_beams()
            for chord in self.get_children():
                chord.finalize()

        self._finalized = True

    def get_subdivision(self):
        if self._subdivision is None:
            if not self.is_filled:
                raise BeatNotFullError()
            self._subdivision = get_chord_group_subdivision(self.get_chords())
        return self._subdivision

    def set_subdivision(self, val):
        self._subdivision = val

    def quantize_quarter_durations(self):
        """
        When called the positioning of children will be quantized according to :obj:`~musicscore.quantize.QuantizeMixin.get_possible_subdivisions()`
        This method is called by :obj:`~musicscore.measure.Measure`

        """
        if self.get_possible_subdivisions() and self.get_children():
            if (
                get_chord_group_subdivision(self.get_children())
                in self.get_possible_subdivisions()
            ):
                pass
            else:
                quarter_durations = [
                    chord.quarter_duration for chord in self.get_children()
                ]
                if len([d for d in quarter_durations if d != 0]) > 1:
                    self._change_children_quarter_durations(
                        self._get_quantized_quarter_durations(quarter_durations)
                    )
                    self._remove_zero_quarter_durations()



from unittest import TestCase
from musicscore import Score, Chord
from musicscore.lyrics import Lyrics
from musicscore.tests.util import _generate_xml_lyric
from musicscore.util import _generate_lyrics
from musicxml import XMLLyric, XMLText, XMLElision, XMLSyllabic
from musicxml.exceptions import XMLElementChildrenRequired

test_lyrics_string = [
    "Bla!",
    "Hello World!",
    "Clou-dy day.",
    "Tra-la-la Ja! Tra-ra! Bah!",
    "No-body is per-fect! - - Are they?",
    "No- - - -body is per- - fect! - - Are they?",
]


def chord_lyric_assertions(chords, expected, number=1):
    for xl1, ch in zip(expected, chords):
        if xl1:
            xl2 = ch._xml_lyrics[number - 1]
            try:
                assert xl1.to_string() == xl2.to_string()
            except XMLElementChildrenRequired:
                xl1.xsd_check = False
                assert xl1.to_string() == xl2.to_string()
        else:
            assert not ch._xml_lyrics


class TestLyrics(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part("p1")

    def test_generate_lyrics_one_word(self):
        xml_lyrics = _generate_lyrics("Bla!")
        expected = _generate_xml_lyric(text="Bla!", syllabic="single")
        assert xml_lyrics[0].to_string() == expected.to_string()

        xml_lyrics = _generate_lyrics(["Bla!"])
        expected = _generate_xml_lyric(text="Bla!", syllabic="single")
        assert xml_lyrics[0].to_string() == expected.to_string()

    def test_generate_lyrics_list_words(self):
        words = ["Hello", "World!"]
        xml_lyrics = _generate_lyrics(words)
        expected = [_generate_xml_lyric(text=t, syllabic="single") for t in words]

        for xl1, xl2 in zip(xml_lyrics, expected):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_two_syllabic_without_None(self):
        word = [("Syl", "lable")]
        expected = [
            _generate_xml_lyric(text="Syl", syllabic="begin"),
            _generate_xml_lyric(text="lable", syllabic="end"),
        ]

        xml_lyrics = _generate_lyrics(word)
        for xl1, xl2 in zip(expected, xml_lyrics):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_syllabic_without_None(self):
        word = [("Clou", "di", "ness")]
        expected = [
            _generate_xml_lyric(text="Clou", syllabic="begin"),
            _generate_xml_lyric(text="di", syllabic="middle"),
            _generate_xml_lyric(text="ness", syllabic="end"),
        ]

        xml_lyrics = _generate_lyrics(word)
        for xl1, xl2 in zip(expected, xml_lyrics):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_list_syllabic_without_None(self):
        words = [("Tra", "la", "la"), "Ja!", ("Tra", "ra!"), "Bah"]

        expected = [
            _generate_xml_lyric(text="Tra", syllabic="begin"),
            _generate_xml_lyric(text="la", syllabic="middle"),
            _generate_xml_lyric(text="la", syllabic="end"),
            _generate_xml_lyric(text="Ja!", syllabic="single"),
            _generate_xml_lyric(text="Tra", syllabic="begin"),
            _generate_xml_lyric(text="ra!", syllabic="end"),
            _generate_xml_lyric(text="Bah", syllabic="single"),
        ]

        xml_lyrics = _generate_lyrics(words)

        for xl1, xl2 in zip(expected, xml_lyrics):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_syllabic_with_None_in_between(self):
        word = ("per", None, None, "fect!")

        expected = [
            _generate_xml_lyric(text="per", syllabic="begin"),
            None,
            None,
            _generate_xml_lyric(text="fect!", syllabic="end"),
        ]

        xml_lyrics = _generate_lyrics(word)

        for xl1, xl2 in zip(expected, xml_lyrics):
            if xl1:
                assert xl1.to_string() == xl2.to_string()
            else:
                assert xl1 == xl2 is None

    def test_generate_lyrics_single_syllabic_with_None_at_the_end(self):
        word = ("ja!", None)

        expected = [
            _generate_xml_lyric(text="ja!", syllabic="single", extend="start"),
            _generate_xml_lyric(extend="stop"),
        ]

        xml_lyrics = _generate_lyrics(word)

        for xl1, xl2 in zip(expected, xml_lyrics):
            try:
                assert xl1.to_string() == xl2.to_string()
            except XMLElementChildrenRequired:
                xl1.xsd_check = False
                assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_syllabic_with_multiple_None_at_the_end(self):
        word = ("per", "fect!", None, None)

        expected = [
            _generate_xml_lyric(text="per", syllabic="begin"),
            _generate_xml_lyric(text="fect!", syllabic="end", extend="start"),
            _generate_xml_lyric(extend="continue"),
            _generate_xml_lyric(extend="stop"),
        ]

        xml_lyrics = _generate_lyrics(word)

        for xl1, xl2 in zip(expected, xml_lyrics):
            try:
                assert xl1.to_string() == xl2.to_string()
            except XMLElementChildrenRequired:
                xl1.xsd_check = False
                assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_list_syllabic_with_None_in_between_and_at_the_end(self):
        words = [
            ("No", None, None, "body"),
            "is",
            ("per", None, None, "fect!", None, None),
            "Are",
            "they?",
        ]
        expected = [
            _generate_xml_lyric(text="No", syllabic="begin"),
            None,
            None,
            _generate_xml_lyric(text="body", syllabic="end"),
            _generate_xml_lyric(text="is", syllabic="single"),
            _generate_xml_lyric(text="per", syllabic="begin"),
            None,
            None,
            _generate_xml_lyric(text="fect!", syllabic="end", extend="start"),
            _generate_xml_lyric(extend="continue"),
            _generate_xml_lyric(extend="stop"),
            _generate_xml_lyric(text="Are", syllabic="single"),
            _generate_xml_lyric(text="they?", syllabic="single"),
        ]

        xml_lyrics = _generate_lyrics(words)

        for xl1, xl2 in zip(expected, xml_lyrics):
            if xl1:
                try:
                    assert xl1.to_string() == xl2.to_string()
                except XMLElementChildrenRequired:
                    xl1.xsd_check = False
                    assert xl1.to_string() == xl2.to_string()
            else:
                assert xl1 == xl2 is None

    def test_generate_lyrics_list_syllabic_with_None_in_between_and_at_the_end_add_to_chords(
        self,
    ):
        words = [
            ("No", None, None, "body"),
            "is",
            ("per", None, None, "fect!", None, None),
            "Are",
            "they?",
        ]
        expected = [
            _generate_xml_lyric(text="No", syllabic="begin"),
            None,
            None,
            _generate_xml_lyric(text="body", syllabic="end"),
            _generate_xml_lyric(text="is", syllabic="single"),
            _generate_xml_lyric(text="per", syllabic="begin"),
            None,
            None,
            _generate_xml_lyric(text="fect!", syllabic="end", extend="start"),
            _generate_xml_lyric(extend="continue"),
            _generate_xml_lyric(extend="stop"),
            _generate_xml_lyric(text="Are", syllabic="single"),
            _generate_xml_lyric(text="they?", syllabic="single"),
        ]

        chords = [Chord(60, 1) for _ in range(len(expected))]
        Lyrics(words).add_to_chords(chords)

        chord_lyric_assertions(chords, expected)

    def test_generate_lyrics_list_syllabic_multiple(self):
        lyrics_1 = [("syl", "labic")]
        lyrics_2 = [("SYL", "LABIC")]
        expected_1 = [
            _generate_xml_lyric(text="syl", syllabic="begin"),
            _generate_xml_lyric(text="labic", syllabic="end"),
        ]
        expected_2 = [
            _generate_xml_lyric(text="SYL", syllabic="begin", number=2),
            _generate_xml_lyric(text="LABIC", syllabic="end", number=2),
        ]
        chords = [Chord(60, 1) for _ in range(2)]
        Lyrics(lyrics_1).add_to_chords(chords)
        Lyrics(lyrics_2, number=2).add_to_chords(chords)
        chord_lyric_assertions(chords, expected_1)
        chord_lyric_assertions(chords, expected_2, 2)

    def test_generate_lyrics_list_syllabic_multiple_show_number(self):
        lyrics_1 = [("syl", "labic")]
        lyrics_2 = [("SYL", "LABIC")]
        first_syllable_1 = XMLLyric(number="1")
        first_syllable_1.add_child(XMLSyllabic("single"))
        first_syllable_1.add_child(XMLText("1."))
        first_syllable_1.add_child(XMLElision(" "))
        first_syllable_1.add_child(XMLSyllabic("begin"))
        first_syllable_1.add_child(XMLText("syl"))

        first_syllable_2 = XMLLyric(number="2")
        first_syllable_2.add_child(XMLSyllabic("single"))
        first_syllable_2.add_child(XMLText("2."))
        first_syllable_2.add_child(XMLElision(" "))
        first_syllable_2.add_child(XMLSyllabic("begin"))
        first_syllable_2.add_child(XMLText("SYL"))

        expected_1 = [
            first_syllable_1,
            _generate_xml_lyric(text="labic", syllabic="end"),
        ]
        expected_2 = [
            first_syllable_2,
            _generate_xml_lyric(text="LABIC", syllabic="end", number=2),
        ]
        chords = [Chord(60, 1) for _ in range(2)]
        Lyrics(lyrics_1, show_number=True).add_to_chords(chords)
        Lyrics(lyrics_2, number=2, show_number=True).add_to_chords(chords)
        chord_lyric_assertions(chords, expected_1)
        chord_lyric_assertions(chords, expected_2, 2)

    def test_generate_lyrics_list_with_None(self):
        chords = [Chord(60, 1) for _ in range(8)]
        Lyrics(
            ["There", "are", "two", "pauses!", None, None, "And", "go!"]
        ).add_to_chords(chords)
        assert chords[4].xml_lyrics == []
        assert chords[5].xml_lyrics == []

    def test_generate_lyrics_with_kwargs(self):
        words = [
            ("No", None, None, "body"),
            "is",
            ("per", None, None, "fect!", None, None),
            "Are",
            "they?",
        ]
        expected = [
            _generate_xml_lyric(text="No", syllabic="begin", default_y=-10),
            None,
            None,
            _generate_xml_lyric(text="body", syllabic="end", default_y=-10),
            _generate_xml_lyric(text="is", syllabic="single", default_y=-10),
            _generate_xml_lyric(text="per", syllabic="begin", default_y=-10),
            None,
            None,
            _generate_xml_lyric(
                text="fect!", syllabic="end", extend="start", default_y=-10
            ),
            _generate_xml_lyric(extend="continue", default_y=-10),
            _generate_xml_lyric(extend="stop", default_y=-10),
            _generate_xml_lyric(text="Are", syllabic="single", default_y=-10),
            _generate_xml_lyric(text="they?", syllabic="single", default_y=-10),
        ]

        chords = [Chord(60, 1) for _ in range(len(expected))]
        Lyrics(words, default_y=-10).add_to_chords(chords)
        chord_lyric_assertions(chords, expected)



import inspect

from musicscore import Chord, Score
from unittest import TestCase
from musicxml.xmlelement.xmlelement import XMLSlur

from musicscore.tests.util import generate_path


class TestSlurs(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part("p1")

    def test_slurs_chords(self):
        ch1 = Chord([60, 62], 2)
        ch2 = Chord([70, 65], 2)
        ch3 = Chord(60, 4)
        ch1.add_x(XMLSlur(type="start"))
        ch2.add_x(XMLSlur(type="stop"))
        ch2.add_x(XMLSlur(type="start"))
        ch3.add_x(XMLSlur(type="stop"))
        self.part.add_chord(ch1)
        self.part.add_chord(ch2)
        self.part.add_chord(ch3)
        path = generate_path(inspect.currentframe())
        self.score.export_xml(path)

    def test_slurs_different_line_types(self):
        line_types = ["dashed", "dotted", "solid", "wavy"]
        for lt in line_types:
            ch1 = Chord(72, 2)
            ch1.add_lyric(lt)
            ch2 = Chord(72, 2)
            ch1.add_x(XMLSlur(type="start", line_type=lt))
            ch2.add_x(XMLSlur(type="stop"))
            self.part.add_chord(ch1)
            self.part.add_chord(ch2)
        path = generate_path(inspect.currentframe())
        self.score.export_xml(path)



import copy

from musicscore import Midi, Chord
from musicscore.exceptions import SimpleFormatException
from musicscore.util import dToX, xToD


class SimpleFormat(object):
    """
    It is useful tool to generate list of chords and also do some simple algorithmic changes to it if needed.
    """

    def __init__(
        self, quarter_durations=None, midis=None, default_midi=71, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._default_midi = None
        self._quarter_durations = None
        self._midis = None
        self._set_quarter_durations(quarter_durations)
        self._set_midis(midis)
        self._chords = None

        self.default_midi = default_midi

    # //private methods
    @staticmethod
    def _check_quarter_duration(quarter_durations):
        for quarter_duration in quarter_durations:
            if quarter_duration < 0:
                raise ValueError(
                    "SimpleFormat(): wrong duration {}".format(quarter_duration)
                )

    def _generate_chords(self):
        self._chords = []
        if self._quarter_durations == [] and self._midis == []:
            pass
        else:
            self._chords = []
            for i in range(len(self._quarter_durations) - len(self._midis)):
                self._midis.append(copy.deepcopy(self.default_midi))

            for i in range(len(self._midis) - len(self._quarter_durations)):
                self._quarter_durations.append(1)

            for d, m in zip(self._quarter_durations, self._midis):
                self._chords.append(Chord(m, d))

    def _set_midis(self, midis):
        def _is_value(x):
            try:
                float(x)
                return True
            except (ValueError, TypeError):
                return False

        def _is_midi(x):
            if isinstance(x, Midi):
                return True
            else:
                return False

        if midis is None:
            midis = []
        else:
            try:
                midis = list(midis)
            except TypeError:
                midis = [midis]

        self._midis = []
        for m in midis:
            if _is_midi(m):
                self._midis.append(m)
            elif _is_value(m):
                self._midis.append(Midi(m))
            elif isinstance(m, list) or isinstance(m, tuple):
                tmp = []

                for midi in m:
                    if _is_midi(midi):
                        tmp.append(midi)
                    elif _is_value(midi):
                        tmp.append(Midi(midi))
                    else:
                        raise TypeError
                self._midis.append(tmp)
            else:
                raise TypeError

    def _set_quarter_durations(self, quarter_durations):
        if quarter_durations is None:
            quarter_durations = []
        else:
            try:
                quarter_durations = list(quarter_durations)
            except TypeError:
                quarter_durations = [quarter_durations]

        self._check_quarter_duration(quarter_durations)

        self._quarter_durations = quarter_durations

    # //public properties
    @property
    def chords(self):
        if self._chords is None:
            self._generate_chords()
        return self._chords

    @property
    def default_midi(self):
        """
        If only quarter_durations are provided the midi of chord will be set to default_midi. Default value is 71.
        """
        return self._default_midi

    @default_midi.setter
    def default_midi(self, val):
        if isinstance(val, Midi):
            self._default_midi = val
        else:
            self._default_midi = Midi(val)

    @property
    def quarter_duration(self):
        return sum([chord.quarter_duration for chord in self.chords])

    @property
    def midis(self):
        return [chord.midis for chord in self.chords]

    # //public methods

    # add
    def add_chord(self, chord):
        # generate chords if needed
        self.chords
        self._chords.append(chord)
        return chord

    # get

    def get_chord_at_position(self, position):
        chord_index = None
        for index, (qp, qd) in enumerate(
            zip(self.get_quarter_positions(), self.get_quarter_durations())
        ):
            if qp <= position < qd + qp:
                chord_index = index
        if chord_index is None:
            return None
        return self.chords[chord_index]

    def get_midis(self):
        return [chord.midis for chord in self.chords]

    def get_quarter_durations(self):
        return [chord.quarter_duration for chord in self.chords]

    def get_quarter_positions(self):
        return dToX(self.get_quarter_durations())

    # //other

    def change_chords(self, function):
        """
        The given function will be applied to all chords.
        """
        for chord in self.chords:
            function(chord)

    def extend(self, simple_format):
        """
        Chords of another SimpleFormat will be added at the end of self.chords.
        """
        for chord in simple_format.chords:
            self.add_chord(chord)

    def mirror(self, pivot=None):
        """
        Chords will be changed to get a mirrored version of the original. Pivot is a midi value that will be used as mirror axis. If not set the first midi will be the axis.
        """
        if pivot is None:
            pivot = self.chords[0].midis[0]
        elif not isinstance(pivot, Midi):
            pivot = Midi(pivot)
        else:
            pass

        for midi in [m for ch in self.chords for m in ch.midis]:
            transposition_interval = (pivot.value - midi.value) * 2
            midi.transpose(transposition_interval)

    def multiply_quarter_durations(self, factor):
        for chord in self.chords:
            chord.quarter_duration *= factor

    @staticmethod
    def sum(*simple_formats, no_duplicates=False):
        if len(set([sf.quarter_duration for sf in simple_formats])) > 1:
            raise SimpleFormatException(
                "SimpleFormat.sum() cannot be used on simple_formats with different durations."
            )
        for midi in [
            midi
            for simple_format in simple_formats
            for chord in simple_format.chords
            for midi in chord.midis
        ]:
            if midi._ties != set():
                raise SimpleFormatException(
                    "SimpleFormat.sum() cannot be used on simple_formats containing tied notes."
                )

        def extract_chord_midis(i):
            def remove_duplicates(new_midis):
                def check(m):
                    for nm in newer_midis:
                        if (
                            m.value == nm.value
                            and m.accidental.mode == nm.accidental.mode
                        ):
                            return False
                    return True

                newer_midis = []
                for m in new_midis:
                    if check(m):
                        newer_midis.append(m)

                return newer_midis

            chords = ordered_chords[i]
            output = []
            for chord in chords:
                new_ms = [m.__deepcopy__() for m in chord.midis if m.value != 0]
                try:
                    if chord in ordered_chords[i + 1]:
                        for m in new_ms:
                            m.add_tie("start")
                except IndexError:
                    pass
                try:
                    if chord in ordered_chords[i - 1]:
                        for m in new_ms:
                            m.add_tie("stop")
                except IndexError:
                    pass
                output.extend(new_ms)
            if no_duplicates:
                output = remove_duplicates(output)
            return output

        sf = SimpleFormat()
        combined_positions_and_chords = {}
        for simple_format in simple_formats:
            for qp in simple_format.get_quarter_positions():
                combined_positions_and_chords[qp] = []
        for key in combined_positions_and_chords.keys():
            for simple_format in simple_formats:
                chord = simple_format.get_chord_at_position(key)
                if chord is not None:
                    combined_positions_and_chords[key] += [chord]
        sum_positions = sorted(combined_positions_and_chords.keys())
        sum_quarter_durations = xToD(sum_positions)
        ordered_chords = [
            combined_positions_and_chords[position] for position in sum_positions[:-1]
        ]
        for index, qd in enumerate(sum_quarter_durations):
            midis = extract_chord_midis(index)
            sf.add_chord(Chord(midis, qd))
        return sf

    def retrograde(self):
        self._chords = list(reversed(self.chords))

    def transpose(self, interval_midi):
        if isinstance(interval_midi, Midi):
            interval = interval_midi.value - self.chords[0].midis[0].value
        else:
            interval = interval_midi
        for ch in self.chords:
            ch.transpose(interval)



from musicscore.tests.test_metronome import TestCase
from musicscore.tests.util_subdivisions import (
    generate_all_subdivisions,
    subdivision_patterns_generator,
    get_first_pattern,
    get_next_pattern,
    generate_all_triplets_manually,
    generate_all_16ths_manually,
    generate_all_quintuplets_manually,
    generate_all_sextuplets_manually,
    generate_all_septuplets_manually,
    generate_all_32nds_manually,
)


def _test_list_of_patterns(patterns1, patterns2, show_diff=False):
    if len(patterns1) != len(patterns2):
        if show_diff:
            print(f"different lengths {len(patterns1)} != {len(patterns2)}")
        return False
    a = [sorted(x) for x in patterns1]
    b = [sorted(x) for x in patterns2]
    while a:
        popped = a.pop()
        if popped not in b:
            if show_diff:
                print(f"{popped} not in patterns2")
            return False
    return True


def get_diff(list1, list2):
    diff_1 = []
    diff_2 = []
    for x in list1:
        if x not in list2:
            diff_1.append(x)
    for x in list2:
        if x not in list1:
            diff_2.append(x)
    if not diff_1 and not diff_2:
        return None
    else:
        return {"list1": diff_1, "list_2": diff_2}


class TestGeneratePatterns(TestCase):
    def test_get_first_pattern(self):
        assert get_first_pattern(2, 3) == [2, 1]
        assert get_first_pattern(1, 3) == [1, 1, 1]
        assert get_first_pattern(3, 4) == [3, 1]
        assert get_first_pattern(2, 4) == [2, 2]
        assert get_first_pattern(1, 4) == [1, 1, 1, 1]
        assert get_first_pattern(4, 5) == [4, 1]
        assert get_first_pattern(3, 5) == [3, 2]
        assert get_first_pattern(2, 5) == [2, 2, 1]
        assert get_first_pattern(1, 5) == [1, 1, 1, 1, 1]
        assert get_first_pattern(5, 6) == [5, 1]
        assert get_first_pattern(4, 6) == [4, 2]
        assert get_first_pattern(3, 6) == [3, 3]
        assert get_first_pattern(2, 6) == [2, 2, 2]
        assert get_first_pattern(1, 6) == [1, 1, 1, 1, 1, 1]

    def test_next_pattern_None(self):
        self.assertIsNone(get_next_pattern([1, 1, 1]))

    def test_next_pattern_is_a_first(self):
        assert get_next_pattern([5, 1]) == [4, 2]
        assert get_next_pattern([4, 1, 1]) == [3, 3]
        assert get_next_pattern([3, 1, 1, 1]) == [2, 2, 2]
        assert get_next_pattern([2, 1, 1, 1, 1]) == [1, 1, 1, 1, 1, 1]

    def test_next_pattern_of_first(self):
        assert get_next_pattern([4, 2]) == [4, 1, 1]
        assert get_next_pattern([3, 3]) == [3, 2, 1]
        assert get_next_pattern([2, 2, 2]) == [2, 2, 1, 1]

    def test_next_pattern_of_not_first(self):
        assert get_next_pattern([3, 2, 1]) == [3, 1, 1, 1]
        assert get_next_pattern([2, 2, 1, 1]) == [2, 1, 1, 1, 1]

    def test_all_next_patterns_until_10(self):
        expected = [[2, 1], [1, 1, 1]]
        assert list(subdivision_patterns_generator(3)) == expected
        expected = [[3, 1], [2, 2], [2, 1, 1], [1, 1, 1, 1]]
        assert list(subdivision_patterns_generator(4)) == expected
        expected = [[4, 1], [3, 2], [3, 1, 1], [2, 2, 1], [2, 1, 1, 1], [1, 1, 1, 1, 1]]
        assert list(subdivision_patterns_generator(5)) == expected
        expected = [
            [5, 1],
            [4, 2],
            [4, 1, 1],
            [3, 3],
            [3, 2, 1],
            [3, 1, 1, 1],
            [2, 2, 2],
            [2, 2, 1, 1],
            [2, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
        ]
        assert list(subdivision_patterns_generator(6)) == expected
        expected = [
            [6, 1],
            [5, 2],
            [5, 1, 1],
            [4, 3],
            [4, 2, 1],
            [4, 1, 1, 1],
            [3, 3, 1],
            [3, 2, 2],
            [3, 2, 1, 1],
            [3, 1, 1, 1, 1],
            [2, 2, 2, 1],
            [2, 2, 1, 1, 1],
            [2, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1],
        ]
        assert list(subdivision_patterns_generator(7)) == expected
        expected = [
            [7, 1],
            [6, 2],
            [6, 1, 1],
            [5, 3],
            [5, 2, 1],
            [5, 1, 1, 1],
            [4, 4],
            [4, 3, 1],
            [4, 2, 2],
            [4, 2, 1, 1],
            [4, 1, 1, 1, 1],
            [3, 3, 2],
            [3, 3, 1, 1],
            [3, 2, 2, 1],
            [3, 2, 1, 1, 1],
            [3, 1, 1, 1, 1, 1],
            [2, 2, 2, 2],
            [2, 2, 2, 1, 1],
            [2, 2, 1, 1, 1, 1],
            [2, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1],
        ]
        assert list(subdivision_patterns_generator(8)) == expected
        expected = [
            [8, 1],
            [7, 2],
            [7, 1, 1],
            [6, 3],
            [6, 2, 1],
            [6, 1, 1, 1],
            [5, 4],
            [5, 3, 1],
            [5, 2, 2],
            [5, 2, 1, 1],
            [5, 1, 1, 1, 1],
            [4, 4, 1],
            [4, 3, 2],
            [4, 3, 1, 1],
            [4, 2, 2, 1],
            [4, 2, 1, 1, 1],
            [4, 1, 1, 1, 1, 1],
            [3, 3, 3],
            [3, 3, 2, 1],
            [3, 3, 1, 1, 1],
            [3, 2, 2, 2],
            [3, 2, 2, 1, 1],
            [3, 2, 1, 1, 1, 1],
            [3, 1, 1, 1, 1, 1, 1],
            [2, 2, 2, 2, 1],
            [2, 2, 2, 1, 1, 1],
            [2, 2, 1, 1, 1, 1, 1],
            [2, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1, 1],
        ]
        assert list(subdivision_patterns_generator(9)) == expected
        expected = [
            [9, 1],
            [8, 2],
            [8, 1, 1],
            [7, 3],
            [7, 2, 1],
            [7, 1, 1, 1],
            [6, 4],
            [6, 3, 1],
            [6, 2, 2],
            [6, 2, 1, 1],
            [6, 1, 1, 1, 1],
            [5, 5],
            [5, 4, 1],
            [5, 3, 2],
            [5, 3, 1, 1],
            [5, 2, 2, 1],
            [5, 2, 1, 1, 1],
            [5, 1, 1, 1, 1, 1],
            [4, 4, 2],
            [4, 4, 1, 1],
            [4, 3, 3],
            [4, 3, 2, 1],
            [4, 3, 1, 1, 1],
            [4, 2, 2, 2],
            [4, 2, 2, 1, 1],
            [4, 2, 1, 1, 1, 1],
            [4, 1, 1, 1, 1, 1, 1],
            [3, 3, 3, 1],
            [3, 3, 2, 2],
            [3, 3, 2, 1, 1],
            [3, 3, 1, 1, 1, 1],
            [3, 2, 2, 2, 1],
            [3, 2, 2, 1, 1, 1],
            [3, 2, 1, 1, 1, 1, 1],
            [3, 1, 1, 1, 1, 1, 1, 1],
            [2, 2, 2, 2, 2],
            [2, 2, 2, 2, 1, 1],
            [2, 2, 2, 1, 1, 1, 1],
            [2, 2, 1, 1, 1, 1, 1, 1],
            [2, 1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        ]
        assert list(subdivision_patterns_generator(10)) == expected

    def test_test_list_patterns(self):
        expected = [[1, 1, 1], [1, 2]]
        a = expected
        b = [[1, 2], [1, 1, 1]]
        c = [[1, 1, 1], [2, 1]]
        d = [[2, 1], [1, 1, 1]]
        e = [[1, 1, 1]]
        f = expected + [[2, 2]]

        assert _test_list_of_patterns(expected, a)
        assert _test_list_of_patterns(expected, b)
        assert _test_list_of_patterns(expected, c, show_diff=True)
        assert _test_list_of_patterns(expected, d)
        assert not _test_list_of_patterns(expected, e)
        assert not _test_list_of_patterns(expected, f)


class TestGenerateAllSubdivisions(TestCase):
    def test_generate_all_triplets(self):
        self.assertIsNone(
            get_diff(generate_all_subdivisions(3), generate_all_triplets_manually())
        )

    def test_generate_all_16ths(self):
        self.assertIsNone(
            get_diff(
                generate_all_subdivisions(4, remove_larger_subdivisions=True),
                generate_all_16ths_manually(),
            )
        )

    def test_generate_all_quintuplets(self):
        self.assertIsNone(
            get_diff(generate_all_subdivisions(5), generate_all_quintuplets_manually())
        )

    def test_generate_all_sextuplets(self):
        self.assertIsNone(
            get_diff(
                generate_all_subdivisions(6, remove_larger_subdivisions=True),
                generate_all_sextuplets_manually(),
            )
        )

    def test_generate_all_septuplets(self):
        self.assertIsNone(
            get_diff(generate_all_subdivisions(7), generate_all_septuplets_manually())
        )

    def test_generate_all_32nds(self):
        self.assertIsNone(
            get_diff(
                generate_all_subdivisions(8, remove_larger_subdivisions=True),
                generate_all_32nds_manually(),
            )
        )

    def test_number_of_patterns(self):
        output = [len(list(subdivision_patterns_generator(x))) for x in range(2, 16)]
        expected = [1, 2, 4, 6, 10, 14, 21, 29, 41, 55, 76, 100, 134, 175]
        assert expected == output



import itertools
import math

from fractions import Fraction


def _permute(l):
    return list(dict.fromkeys(itertools.permutations(l)))


def generate_all_triplets_manually():
    output = [tuple(3 * [Fraction(1, 3)])]
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 3), Fraction(2, 3)])))
    )
    return output


def generate_all_16ths_manually():
    output = [tuple(4 * [Fraction(1, 4)])]
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 4), Fraction(1, 4), Fraction(1, 2)])
            )
        )
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 4), Fraction(3, 4)])))
    )
    return output


def generate_all_quintuplets_manually():
    output = [tuple(5 * [Fraction(1, 5)])]
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 5), Fraction(1, 5), Fraction(1, 5), Fraction(2, 5)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 5), Fraction(2, 5), Fraction(2, 5)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 5), Fraction(1, 5), Fraction(3, 5)])
            )
        )
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(2, 5), Fraction(3, 5)])))
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 5), Fraction(4, 5)])))
    )
    return output


def generate_all_sextuplets_manually():
    output = [tuple(6 * [Fraction(1, 6)])]
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [
                        Fraction(1, 6),
                        Fraction(1, 6),
                        Fraction(1, 6),
                        Fraction(1, 6),
                        Fraction(2, 6),
                    ]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 6), Fraction(1, 6), Fraction(2, 6), Fraction(2, 6)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(3, 6)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 6), Fraction(2, 6), Fraction(3, 6)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 6), Fraction(1, 6), Fraction(4, 6)])
            )
        )
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 6), Fraction(5, 6)])))
    )
    return output


def generate_all_septuplets_manually():
    output = [tuple(7 * [Fraction(1, 7)])]
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(2, 7),
                    ]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(2, 7),
                        Fraction(2, 7),
                    ]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 7), Fraction(2, 7), Fraction(2, 7), Fraction(2, 7)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(3, 7),
                    ]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 7), Fraction(1, 7), Fraction(2, 7), Fraction(3, 7)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 7), Fraction(3, 7), Fraction(3, 7)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(4, 7)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 7), Fraction(2, 7), Fraction(4, 7)])
            )
        )
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(3, 7), Fraction(4, 7)])))
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(3, 7), Fraction(2, 7), Fraction(2, 7)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(5, 7)])
            )
        )
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(2, 7), Fraction(5, 7)])))
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(6, 7)])))
    )
    return output


def generate_all_32nds_manually():
    output = [tuple(8 * [Fraction(1, 8)])]
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(1 * [Fraction(2, 8)] + 6 * [Fraction(1, 8)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(2 * [Fraction(2, 8)] + 4 * [Fraction(1, 8)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(3 * [Fraction(2, 8)] + 2 * [Fraction(1, 8)])
            )
        )
    )

    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(1 * [Fraction(3, 8)] + 5 * [Fraction(1, 8)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    1 * [Fraction(3, 8)] + 1 * [Fraction(2, 8)] + 3 * [Fraction(1, 8)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    1 * [Fraction(3, 8)] + 2 * [Fraction(2, 8)] + 1 * [Fraction(1, 8)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(2 * [Fraction(3, 8)] + 2 * [Fraction(1, 8)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(2 * [Fraction(3, 8)] + 1 * [Fraction(2, 8)])
            )
        )
    )

    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(1 * [Fraction(4, 8)] + 4 * [Fraction(1, 8)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    1 * [Fraction(4, 8)] + 1 * [Fraction(2, 8)] + 2 * [Fraction(1, 8)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    1 * [Fraction(4, 8)] + 1 * [Fraction(3, 8)] + 1 * [Fraction(1, 8)]
                )
            )
        )
    )

    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(1 * [Fraction(5, 8)] + 3 * [Fraction(1, 8)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    1 * [Fraction(5, 8)] + 1 * [Fraction(2, 8)] + 1 * [Fraction(1, 8)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(1 * [Fraction(5, 8)] + 1 * [Fraction(3, 8)])
            )
        )
    )

    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(1 * [Fraction(6, 8)] + 2 * [Fraction(1, 8)])
            )
        )
    )

    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(1 * [Fraction(7, 8)] + 1 * [Fraction(1, 8)])
            )
        )
    )
    return output


def generate_all_nonuplets():
    output = [tuple(9 * [Fraction(1, 9)])]
    # proportions = [
    #     1 * [2] + 7 * [1],
    #     2 * [2] + 5 * [1],
    #     3 * [1] + 3 * [2],
    #     1 * [1] + 4 * [2],
    #     6 * [1] + [3],
    #     3 * [1] + 2 * [3],
    #     3 * [3],
    #     5 * [1] + [4],
    #     4 * [1] + [5],
    #     3 * [1] + [6],
    #     2 * [1] + [7],
    #     1 * [1] + [8],
    # ]
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(2, 7),
                    ]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(2, 7),
                        Fraction(2, 7),
                    ]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 7), Fraction(2, 7), Fraction(2, 7), Fraction(2, 7)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(1, 7),
                        Fraction(3, 7),
                    ]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 7), Fraction(1, 7), Fraction(2, 7), Fraction(3, 7)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 7), Fraction(3, 7), Fraction(3, 7)])
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations(
                    [Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(4, 7)]
                )
            )
        )
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 7), Fraction(2, 7), Fraction(4, 7)])
            )
        )
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(3, 7), Fraction(4, 7)])))
    )
    output.extend(
        list(
            dict.fromkeys(
                itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(5, 7)])
            )
        )
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(2, 7), Fraction(5, 7)])))
    )
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(6, 7)])))
    )
    return output


def get_first_pattern(first_element, sum_):
    output = (sum_ // first_element) * [first_element]
    if sum(output) != sum_:
        output += [sum_ - sum(output)]
    return output


def get_next_pattern(current_pattern):
    if set(current_pattern) == {1}:
        return None
    subdivision = sum(current_pattern)
    first_element = current_pattern[0]
    if len(current_pattern) > 1:
        other_elements = current_pattern[1:]
    else:
        other_elements = None
    if not other_elements or set(other_elements) == {1}:
        return get_first_pattern(first_element - 1, subdivision)
    else:
        output = [first_element] + get_next_pattern(other_elements)
        return output


def subdivision_patterns_generator(subdivision):
    current = [subdivision]
    current = get_next_pattern(current)
    while current:
        yield current
        current = get_next_pattern(current)


def generate_all_subdivision_patterns(subdivision, remove_larger_subdivisions=False):
    output = []
    patterns = subdivision_patterns_generator(subdivision)
    if remove_larger_subdivisions:
        patterns = [
            pat for pat in patterns if not (min(pat) != 1 and math.gcd(*pat) != 1)
        ]
    for pattern in patterns:
        output.extend(_permute(pattern))
    return output


def generate_all_subdivisions(subdivision, remove_larger_subdivisions=False):
    return [
        tuple([Fraction(x, subdivision) for x in pat])
        for pat in generate_all_subdivision_patterns(
            subdivision, remove_larger_subdivisions
        )
    ]


def generate_subdivsion_test_patterns(subdivision, remove_larger_subdivision=False):
    def get_last_element():
        return subdivision - x - offset

    offset = 0
    while offset < subdivision:
        x = 1
        while get_last_element():
            if offset:
                pattern = (offset, x, get_last_element())
            else:
                pattern = (x, get_last_element())
            x += 1
            if remove_larger_subdivision and (
                min(pattern) != 1 and math.gcd(*pattern) != 1
            ):
                continue
            yield pattern
        offset += 1



import math
import xml.etree.ElementTree as ET
from pathlib import Path

import xmltodict

from musicscore import Time, SimpleFormat, BassClef, TrebleClef
from musicscore.chord import Chord
from musicscore.exceptions import (
    ScorePartIdIsNotUniqueError,
)
from musicscore.key import Key
from musicscore.measure import Measure
from musicscore.part import Part, ScorePart
from musicscore.quarterduration import QuarterDuration
from musicscore.score import Score
from musicscore.tests.test_metronome import TestCase
from musicscore.tests.util import (
    get_xml_elements_diff,
    XMLsDifferException,
    get_xml_diff_part,
    generate_xml_file,
    create_test_xml_paths,
)

path = Path(__file__)


class TestPart(TestCase):
    def test_part_init(self):
        p = Part(id="p1")
        p.add_child(Measure(1))
        assert p.xml_object.id == "p1"

    def test_part_name(self):
        p = Part(id="p1")
        assert p.name == ""
        p = Part(id="p2", name="Part 1")
        assert p.name == "Part 1"
        p.name = None
        assert p.name == ""

    def test_part_abbreviation(self):
        p = Part(id="p1")
        assert p.abbreviation is None
        p = Part(id="p2", name="Part 1", abbreviation="p 1")
        assert p.abbreviation == "p 1"
        p = Part(id="p3", name="Part 3")
        p.abbreviation = "p 3"
        assert p.abbreviation == "p 3"

    def test_part_and_score_part(self):
        p = Part(id="p1")
        assert isinstance(p.score_part, ScorePart)
        assert p.score_part.xml_object.id == p.xml_object.id

    def test_part_list_multiple_parts(self):
        score = Score()
        score.add_part("p1")
        score.add_part("p2")
        score.finalize()
        assert len(score.xml_object.xml_part_list.get_children()) == 2

    def test_add_measure(self):
        p = Part("p1")
        m = p.add_measure()
        assert p.get_children()[-1] == m
        assert m.number == 1
        assert m.time.signatures == (4, 4)
        m = p.add_measure(time=(4, 3, 2, 1))
        assert p.get_children()[-1] == m
        assert m.number == 2
        assert m.time.signatures == (4, 3, 2, 1)
        m = p.add_measure()
        assert p.get_children()[-1] == m
        assert m.number == 3
        assert m.time.signatures == (4, 3, 2, 1)

    def test_add_measure_show_time(self):
        p = Part("p1")
        m = p.add_measure()
        assert m.time.show is True
        m = p.add_measure()
        assert m.time.show is False
        m = p.add_measure(time=(3, 4))
        assert m.time.show is True

    def test_add_measure_show_key(self):
        p = Part("p1")
        m = p.add_measure()
        assert m.key.show is True
        m.finalize()
        assert m.xml_object.xml_attributes.xml_key is not None
        m = p.add_measure()
        m.finalize()
        assert m.key.show is False
        m.key = Key(fifths=1)
        assert m.key.show is True
        m = p.add_measure()
        m.finalize()
        assert m.key.fifths == 1
        assert m.key.show is False
        assert m.xml_object.xml_attributes.xml_key is None

    def test_add_measure_show_clefs(self):
        p = Part("p1")
        m = p.add_measure()
        m.add_staff(1)
        m.add_staff(2)
        assert m.clefs[0].show is True
        assert m.clefs[1].show is True
        m = p.add_measure()
        assert m.clefs[0].show is False
        assert m.clefs[1].show is False
        m = p.add_measure()
        m.clefs[0].show = True
        m.finalize()
        clefs = m.xml_object.xml_attributes.find_children("XMLClef")
        assert len(clefs) == 1
        assert clefs[0].xml_sign.value_ == "G"
        assert clefs[0].xml_line.value_ == 2

    def test_part_add_measure_check_voice(self):
        """
        Test if Part.add_measure() adds a Measure with a Staff (number=None) and Voice (number=1)
        """
        p = Part("p1")
        m1 = p.add_measure()
        assert m1.get_staff(1) is not None
        assert m1.get_voice(staff_number=1, voice_number=1) is not None
        m1.add_staff(2)
        m2 = p.add_measure()
        m2.add_voice(staff_number=2, voice_number=2)

        assert m1.get_voice(staff_number=2, voice_number=2) is None
        assert m2.get_voice(staff_number=2, voice_number=2) is not None

    def test_part_set_current_measure(self):
        p = Part("p1")
        m1 = Measure(1)
        p.set_current_measure(1, 1, m1)
        assert p.get_current_measure(1, 1) == m1
        m2 = Measure(2)
        p.set_current_measure(1, 1, m2)
        assert p.get_current_measure(1, 1) == m2
        assert p.get_current_measure(1, 2) is None

        p = Part("p2")
        m1 = Measure(1)
        p.set_current_measure(None, 1, m1)
        assert p.get_current_measure(1, 1) == m1
        m2 = Measure(2)
        p.set_current_measure(1, 1, m2)
        assert p.get_current_measure(1, 1) == m2

    def test_part_get_current_measure_simple(self):
        p = Part("p1")
        m1 = p.add_measure()
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        m2 = p.add_measure()
        m1._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        m1._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        assert m1.get_voice(staff_number=1, voice_number=1).is_filled
        m2._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m2
        _ = p.add_measure()
        m4 = p.add_measure()
        m4._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m4

    def test_part_get_current_measure_complex(self):
        p = Part("p1")
        assert p.get_current_measure(staff_number=None, voice_number=1) is None
        m1 = p.add_measure()
        m1._add_chord(
            Chord(midis=60, quarter_duration=4), staff_number=2, voice_number=2
        )
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        assert p.get_current_measure(staff_number=1, voice_number=2) is None
        assert p.get_current_measure(staff_number=2, voice_number=1) == m1
        assert (
            p.get_current_measure(staff_number=2, voice_number=1)
            .get_voice(staff_number=2, voice_number=1)
            .is_filled
            is False
        )
        assert p.get_current_measure(staff_number=2, voice_number=2) == m1
        assert (
            p.get_current_measure(staff_number=2, voice_number=2)
            .get_voice(staff_number=2, voice_number=2)
            .is_filled
            is True
        )
        m1.add_voice(staff_number=1, voice_number=1)
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        #
        m2 = p.add_measure()
        m2._add_chord(
            Chord(midis=60, quarter_duration=4), staff_number=2, voice_number=2
        )

        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        assert p.get_current_measure(staff_number=1, voice_number=2) is None
        assert p.get_current_measure(staff_number=2, voice_number=1) == m1
        assert p.get_current_measure(staff_number=2, voice_number=2) == m2

        m1 = Measure(1)
        m1._add_chord(Chord(midis=60, quarter_duration=1))
        p = Part("p2")
        p.add_child(m1)
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1


class TestScorePart(TestCase):
    def test_score_part_name(self):
        p = Part(id="p1")
        assert p.score_part.xml_part_name.value_ == p.name == ""
        p.name = "Part 1"
        assert p.score_part.xml_part_name.value_ == p.name == "Part 1"
        p.name = None
        assert p.score_part.xml_part_name.value_ == p.name == ""

    def test_score_part_abbreviation(self):
        p = Part(id="p1")
        assert p.score_part.xml_part_abbreviation == p.abbreviation is None
        p = Part(id="p2", name="Part 1", abbreviation="p 1")
        assert p.score_part.xml_part_abbreviation.value_ == p.abbreviation == "p 1"
        p = Part(id="p3", name="Part 3")
        p.abbreviation = "p 3"
        assert p.score_part.xml_part_abbreviation.value_ == p.abbreviation == "p 3"

    def test_score_part_id(self):
        p = Part(id="p1")
        assert p.score_part.id == p.score_part.xml_object.id == p.id_ == "p1"
        p.id_ = "p2"
        assert p.score_part.id == p.score_part.xml_object.id == p.id_ == "p2"
        p.score_part.to_string()
        expected = """<score-part id="p2">
  <part-name />
</score-part>
"""
        assert p.score_part.to_string() == expected

    def test_score_part_to_string(self):
        p = Part(id="p1")
        expected = """<score-part id="p1">
  <part-name />
</score-part>
"""
        assert p.score_part.to_string() == expected

    def test_add_measure_clef(self):
        p = Part("P1")
        m1 = p.add_measure(time=(3, 4))
        m2 = p.add_measure(time=(2, 4))
        p.finalize()
        assert m1.clefs[0].show is True
        assert m2.clefs[0].show is False

    def test_add_complex_quarter_durations(self):
        s = Score()
        p = s.add_child(Part("p1"))
        # quarter_durations = [
        #     QuarterDuration(145, 42),
        #     QuarterDuration(1070, 737),
        #     QuarterDuration(1121, 352),
        #     QuarterDuration(960, 589),
        #     QuarterDuration(178, 85),
        #     QuarterDuration(134, 945),
        #     QuarterDuration(1733, 482),
        #     QuarterDuration(446, 227),
        #     QuarterDuration(458, 735),
        #     QuarterDuration(943, 890),
        #     QuarterDuration(650, 739),
        #     QuarterDuration(939, 338),
        #     QuarterDuration(327, 227),
        #     QuarterDuration(461, 183),
        #     QuarterDuration(1, 2),
        #     QuarterDuration(1123, 457),
        #     QuarterDuration(24, 59),
        #     QuarterDuration(1075, 397),
        #     QuarterDuration(604, 219),
        #     QuarterDuration(18, 605),
        #     QuarterDuration(2398, 999),
        #     QuarterDuration(235, 608),
        #     QuarterDuration(677, 799),
        #     QuarterDuration(136, 637),
        #     QuarterDuration(2717, 783),
        #     QuarterDuration(5, 643),
        # ]
        # first_qd = math.ceil(sum(quarter_durations[:4])) - sum(quarter_durations[:4])
        # qds = [first_qd] + [quarter_durations[4]]
        # last_qd = math.ceil(sum(qds)) - sum(qds)
        # qds += [last_qd]
        # qds = quarter_durations[:5]
        # last_qd = math.ceil(sum(qds)) - sum(qds)
        # qds += [last_qd]
        # qds = [QuarterDuration(145, 42), QuarterDuration(1070, 737), QuarterDuration(1121, 352), QuarterDuration(960, 589),
        #        QuarterDuration(178, 85), QuarterDuration(32, 171)]
        # qds = qds[4:]
        # first_qd = math.ceil(sum(qds)) - sum(qds)
        # print(first_qd)
        # qds = [first_qd] + qds
        # print(qds)
        qds = [
            QuarterDuration(23, 32),
            QuarterDuration(178, 85),
            QuarterDuration(32, 171),
            QuarterDuration(1),
        ]
        for qd in qds:
            p.add_chord(Chord(midis=60, quarter_duration=qd))
        expected = [
            QuarterDuration(23, 32),
            QuarterDuration(9, 32),
            QuarterDuration(1, 1),
            QuarterDuration(139, 171),
            QuarterDuration(32, 171),
            QuarterDuration(1, 1),
        ]

        assert [ch.quarter_duration for ch in p.get_chords()] == expected

        for beat in [
            b
            for m in p.get_children()
            for st in m.get_children()
            for v in st.get_children()
            for b in v.get_children()
        ]:
            beat.quantize_quarter_durations()
            beat._split_not_writable_chords()
        expected = [
            QuarterDuration(3, 7),
            QuarterDuration(2, 7),
            QuarterDuration(2, 7),
            QuarterDuration(1, 1),
            QuarterDuration(4, 5),
            QuarterDuration(1, 5),
            QuarterDuration(1, 1),
        ]
        assert [ch.quarter_duration for ch in p.get_chords()] == expected

    def test_chord_previous(self):
        p = Part("P1")
        for qd in [1, 2.5, 0.5]:
            p.add_chord(Chord(midis=60, quarter_duration=qd))

        for i in range(1, len(p.get_children())):
            current = p.get_children()[i]
            previous = p.get_children()[i - 1]
            assert current.previous == previous

    def test_chord_offsets(self):
        p = Part("P1")
        for qd in [1, 2.5, 0.5]:
            p.add_chord(Chord(midis=60, quarter_duration=qd))

        assert [ch.quarter_duration for ch in p.get_chords()] == [1, 2, 0.5, 0.5]
        assert [ch.offset for ch in p.get_chords()] == [0, 0, 0, 0.5]


class TestAddChordToPart(TestCase):
    def setUp(self):
        self.score = Score()
        super().setUp()

    @staticmethod
    def _get_clefs_of_measure(part, measure_number=1):
        return xmltodict.parse(part.get_children()[measure_number - 1].to_string())[
            "measure"
        ]["attributes"].get("clef", None)

    def test_add_one_chord_quarter_duration_4(self):
        xml_file = create_test_xml_paths(path, "add_one_chord_quarter_duration_4")[0]
        p = self.score.add_part(id="part1")
        chord = Chord(60, 4)
        p.add_chord(chord)
        self.score.export_xml(xml_file)
        output_score = ET.parse(xml_file)
        output_part_xml = output_score.find("part")
        output_part_xml_dict = xmltodict.parse(ET.tostring(output_part_xml))
        expected = {
            "part": {
                "@id": "part1",
                "measure": {
                    "@number": "1",
                    "attributes": {
                        "divisions": "1",
                        "key": {"fifths": "0"},
                        "time": {"beats": "4", "beat-type": "4"},
                        "clef": {"sign": "G", "line": "2"},
                    },
                    "note": {
                        "pitch": {"step": "C", "octave": "4"},
                        "duration": "4",
                        "voice": "1",
                        "type": "whole",
                    },
                    "barline": {"@location": "right", "bar-style": "light-heavy"},
                },
            }
        }
        assert output_part_xml_dict == expected

    def test_add_chord_to_staff_with_bass_clef(self):
        part = Part(id="part-1")
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {
            "@number": "2",
            "sign": "F",
            "line": "4",
        }
        assert self._get_clefs_of_measure(part, 2) is None

    def test_add_chord_to_both_staves(self):
        part = Part(id="part-1")
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {
            "@number": "2",
            "sign": "F",
            "line": "4",
        }
        assert self._get_clefs_of_measure(part, 2) is None

    def test_add_chord_to_both_staves_other_order(self):
        part = Part(id="part-1")
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)

        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)

        assert self._get_clefs_of_measure(part, 1)[1] == {
            "@number": "2",
            "sign": "F",
            "line": "4",
        }
        assert self._get_clefs_of_measure(part, 2) is None

    def test_add_long_chord_to_staff_with_bass_clef(self):
        part = Part(id="part-1")
        part.add_chord(Chord(midis=60, quarter_duration=11), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {
            "@number": "2",
            "sign": "F",
            "line": "4",
        }
        assert self._get_clefs_of_measure(part, 2) is None
        assert self._get_clefs_of_measure(part, 3) is None

    def test_different_long_chords_to_staff_with_bass_clef(self):
        part = Part(id="part-1")
        midis = list(range(60, 65))
        quarter_durations = list(range(1, 6))
        for m, q in zip(midis, quarter_durations):
            part.add_chord(Chord(m, q), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {
            "@number": "2",
            "sign": "F",
            "line": "4",
        }
        for measure_number in range(2, math.ceil(sum(quarter_durations) / 4)):
            assert self._get_clefs_of_measure(part, measure_number) is None

    def test_add_chord_to_two_staves_same_length(self):
        part = Part(id="part-1")
        midis = list(range(60, 65))
        quarter_durations = list(range(1, 6))
        for m, q in zip(midis, quarter_durations):
            part.add_chord(Chord(m, q), staff_number=1)
        for m, q in zip(reversed(midis), reversed(quarter_durations)):
            part.add_chord(Chord(m, q), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {
            "@number": "2",
            "sign": "F",
            "line": "4",
        }
        for measure_number in range(2, math.ceil(sum(quarter_durations) / 4)):
            assert self._get_clefs_of_measure(part, measure_number) is None

    def test_add_chord_different_staves_and_voices(self):
        p = Part("p1")
        p.add_chord(Chord(60, 1))
        m = p.get_children()[0]
        assert m.get_voice(staff_number=1, voice_number=1) is not None
        p.add_chord(Chord(61, 2), staff_number=2, voice_number=4)
        for i in range(1, 5):
            assert m.get_voice(staff_number=2, voice_number=i) is not None
        p.add_chord(Chord(62, 3), staff_number=1, voice_number=1)
        assert [
            ch.quarter_duration
            for ch in m.get_voice(staff_number=1, voice_number=1).get_chords()
        ] == [1, 3]
        assert [
            ch.midis[0].value
            for ch in m.get_voice(staff_number=1, voice_number=1).get_chords()
        ] == [60, 62]
        assert [
            ch.quarter_duration
            for ch in m.get_voice(staff_number=2, voice_number=4).get_chords()
        ] == [2]

    def test_add_chord_with_leftover(self):
        p = Part("p1")
        ch = Chord(60, 5)
        p.add_chord(ch)
        assert len(p.get_children()) == 2
        m1, m2 = p.get_children()
        assert p.get_current_measure(1, 1) == m2
        assert m1.get_voice(staff_number=1, voice_number=1).is_filled
        assert not m2.get_voice(staff_number=1, voice_number=1).is_filled

    def test_add_chord_to_full_measure(self):
        p = Part("p1")
        p.add_chord(Chord(60, 4))
        p.add_chord(Chord(60, 6))
        m1, m2, m3 = p.get_children()
        assert p.get_current_measure() == m3

    def test_add_chord_to_existing_measures(self):
        p = Part("p1")
        for _ in range(3):
            p.add_measure()
        m1, m2, m3 = p.get_children()
        assert p.get_current_measure() == m1
        p.add_chord(Chord(60, 5))
        assert p.get_current_measure() == m2

    def test_add_chord_with_staff_number(self):
        p = Part("P1")
        ch1 = Chord(60, 1)
        ch2 = Chord(48, 1)
        p.add_chord(ch1, staff_number=1)
        assert ch1.get_staff_number() is None
        p.add_chord(ch2, staff_number=2)
        assert ch1.get_staff_number() == 1
        assert ch2.get_staff_number() == 2
        p.finalize()
        assert ch1.notes[0].xml_staff.value_ == 1
        assert ch2.notes[0].xml_staff.value_ == 2

    def test_add_chord_with_clef(self):
        xml_path, expected_path = create_test_xml_paths(path, "add_chord_with_clef")
        sf1 = SimpleFormat(midis=[60, (61, 66), 62], quarter_durations=[1, 2, 3])
        sf2 = SimpleFormat(midis=[60, (61, 66), 62], quarter_durations=[1, 2, 3])
        sf1.chords[1].clef = BassClef()

        sf2.chords[1].clef = BassClef()
        sf2.chords[2].clef = TrebleClef()
        score = Score()
        part = score.add_part(id="part-1")
        for index, simpleformat in enumerate([sf1, sf2]):
            for chord in simpleformat.chords:
                part.add_chord(chord, staff_number=index + 1)
        part.get_staff(1, 2).clef = TrebleClef()
        score.export_xml(xml_path)

        get_xml_diff_part(expected_path, xml_path, Path(__file__))

    def test_add_first_chord_with_clef(self):
        xml_path, expected_path = create_test_xml_paths(
            path, "add_first_chord_with_clef"
        )
        sf1 = SimpleFormat(midis=[60, (61, 66), 62], quarter_durations=[2, 6, 4])
        sf1.chords[0].clef = BassClef()
        sf1.chords[2].clef = TrebleClef()
        generate_xml_file(Score(), sf1, path=xml_path)
        get_xml_diff_part(expected_path, xml_path, Path(__file__))


class TestPartIdUniqueness(TestCase):
    def test_part_id_is_unique(self):
        part_a = Part(id="p1")
        part_b = Part(id="p1")
        score = Score()
        score.add_child(part_a)
        score.add_child(part_b)
        with self.assertRaises(ScorePartIdIsNotUniqueError):
            score.finalize()


class TestSplitQdAndTime(TestCase):
    def test_part_add_split_original_ties_1(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part("p1")
        p.add_measure((5, 4))
        p.add_chord(chord)
        for chord in p.get_chords():
            assert chord._original_starting_ties == [set()]

    def test_part_add_split_original_ties_2(self):
        chord = Chord(midis=[60, 62], quarter_duration=10)
        chord.midis[0].add_tie("start")
        chord.midis[1].add_tie("stop")
        p = Part("p1")
        p.add_measure((5, 4))
        p.add_chord(chord)
        for chord in p.get_chords():
            assert chord._original_starting_ties == [{"start"}, {"stop"}]

    def test_part_add_split_chord_5_3_4(self):
        chord = Chord(midis=60, quarter_duration=5)
        p = Part("p1")
        p.add_measure((3, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 2
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_5_4_4(self):
        chord = Chord(midis=60, quarter_duration=5)
        p = Part("p1")
        p.add_measure((4, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 1
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_5_5_4(self):
        chord = Chord(midis=60, quarter_duration=5)
        p = Part("p1")
        p.add_measure((5, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 1
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [3, 2]
        assert not ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next

    def test_part_add_split_chord_6_4_4(self):
        chord = Chord(midis=60, quarter_duration=6)
        p = Part("p1")
        p.add_measure((4, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 2
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_6_5_4(self):
        chord = Chord(midis=60, quarter_duration=6)
        p = Part("p1")
        p.add_measure((5, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        measure = p.get_children()[-1]
        assert len(measure.get_chords()) == 1
        ch = measure.get_chords()[0]
        assert ch.quarter_duration == 1
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_7_7_4(self):
        chord = Chord(midis=60, quarter_duration=7)
        p = Part("p1")
        p.add_measure((7, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 1
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [4, 3]
        assert not ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next

    def test_part_add_split_chord_10_7_4(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part("p1")
        p.add_measure((7, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 3
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_12_7_4(self):
        chord = Chord(midis=60, quarter_duration=12)
        p = Part("p1")
        p.add_measure((7, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [3, 2]
        assert ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next

    def test_part_add_split_chord_10_5_4(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part("p1")
        p.add_measure((5, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [3, 2]
        assert ch1.all_midis_are_tied_to_next
        assert ch1.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous

    def test_part_add_split_chord_10_10_4(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part("p2")
        p.add_measure((10, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 1
        m = p.get_children()[-1]
        all_chords = [ch1, ch2, ch3] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [4, 4, 2]
        assert not ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert ch2.all_midis_are_tied_to_next
        assert ch3.all_midis_are_tied_to_previous
        assert not ch3.all_midis_are_tied_to_next

    def test_add_chords_with_partially_tied_notes(self):
        xml_path, expected_path = create_test_xml_paths(
            path, "add_chords_with_partially_tied_notes"
        )
        midis = [[60, 63], [61, 63], [62, 64], [62, 65]]
        quarter_durations = [1, 2, 2, 1]
        chords = [Chord(ms, qd) for ms, qd in zip(midis, quarter_durations)]
        chords[0].midis[1].add_tie("start")
        chords[1].midis[1].add_tie("stop")
        chords[2].midis[0].add_tie("start")
        chords[3].midis[0].add_tie("stop")
        p = Part(id="part-1")
        for chord in chords:
            p.add_chord(chord)
        score = Score()
        score.add_child(p)
        score.export_xml(xml_path)
        get_xml_diff_part(expected_path, xml_path, Path(__file__))

    def test_add_chords_with_partially_tied_notes_simplified(self):
        xml_path, expected_path = create_test_xml_paths(
            path, "add_chords_with_partially_tied_notes_simplified"
        )
        midis = [[62, 64], [62, 65]]
        quarter_durations = [3, 1]
        chords = [Chord(ms, qd) for ms, qd in zip(midis, quarter_durations)]
        chords[0].midis[0].add_tie("start")
        chords[1].midis[0].add_tie("stop")
        p = Part("part-1")
        p.add_measure(Time(2, 4))
        for chord in chords:
            p.add_chord(chord)
        score = Score()
        score.add_child(p)
        score.export_xml(xml_path)
        el1 = (
            ET.parse(Path(__file__).parent / xml_path)
            .getroot()
            .find("part[@id='part-1']")
        )
        el2 = (
            ET.parse(Path(__file__).parent / expected_path)
            .getroot()
            .find("part[@id='part-1']")
        )
        diff = get_xml_elements_diff(el1=el1, el2=el2)
        if diff:
            raise XMLsDifferException(diff)



"""
Various clefs: G, C, F, percussion, TAB and none; some are also possible with transposition and on other staff lines than their default (e.g. soprano/alto/tenor/baritone C clefs); Each measure shows a different clef (measure 17 has the " none" clef), only measure 18 has the same treble clef as measure 1.
"""

from pathlib import Path

from musicscore import AltoClef, TenorClef, BassClef, TrebleClef, Score, Chord
from musicscore.clef import PercussionClef, Clef
from unittest import TestCase

clefs = [
    TrebleClef(),
    AltoClef(),
    TenorClef(),
    BassClef(),
    PercussionClef(),
    TrebleClef(octave_change=-1),
    BassClef(octave_change=-1),
    Clef(sign="G", line=1),
    Clef(sign="F", line=3),
    Clef(sign="C", line=5),
    Clef(sign="C", line=2),
    Clef(sign="C", line=1),
    PercussionClef(),
    TrebleClef(octave_change=1),
    BassClef(octave_change=1),
    Clef(sign="TAB"),
    Clef("none"),
    TrebleClef(),
]


class TestLily12a(TestCase):
    def test_lily_12a_Clefs(self):
        score = Score()
        part = score.add_part("p1")
        for c in clefs:
            ch = Chord(60, 4)
            part.add_chord(ch)
            part.get_current_measure().get_staff(1).clef = c

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from unittest import TestCase
from unittest.mock import Mock, patch

from musicscore.beat import Beat
from musicscore.chord import Chord
from musicscore.exceptions import VoiceHasNoBeatsError, AddChordError, VoiceIsFullError
from musicscore.voice import Voice


class TestVoice(TestCase):
    @patch("musicscore.voice.Voice.up", new=Mock())
    def test_voice_add_beat(self):
        v = Voice()
        quarter_durations = [
            1 / 64,
            3 / 64,
            1 / 32,
            3 / 32,
            1 / 16,
            3 / 16,
            1 / 8,
            3 / 8,
            1 / 4,
            3 / 4,
            1 / 2,
            1,
            2,
            3,
        ]
        for quarter_duration in quarter_durations:
            v.add_child(Beat(quarter_duration))
        assert [
            child.quarter_duration for child in v.get_children()
        ] == quarter_durations

    @patch("musicscore.voice.Voice.up")
    def test_update_beats(self, mock_beat):
        v = Voice()
        assert not v.get_children()
        v.update_beats()
        assert not v.get_children()
        v.update_beats(1 / 4, 1 / 4, 1 / 4, 1 / 4)
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 4)] * 4
        v.update_beats([1 / 6] * 6)
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 6)] * 6
        v.update_beats([1 / 4] * 3)
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 4)] * 3

    @patch("musicscore.voice.Voice.up", new=Mock())
    def test_update_beats_with_old_beats(self):
        v = Voice()
        v.update_beats(1 / 4, 1 / 4, 1 / 4, 1 / 4)
        v.update_beats(1 / 4, 1 / 2)

        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 4), (1, 2)]

    @patch("musicscore.voice.Voice.up", new=Mock())
    def test_get_current_beat(self):
        v = Voice()
        beats = v.update_beats(1 / 4, 1 / 4, 1 / 4, 1 / 4)
        assert v.get_current_beat() == beats[0]
        v.get_children()[0]._filled_quarter_duration = 1 / 8
        assert v.get_current_beat() == beats[0]
        v.get_children()[0]._filled_quarter_duration = 1 / 4
        assert v.get_current_beat() == beats[1]
        v.get_children()[1]._filled_quarter_duration = 1 / 8
        assert v.get_current_beat() == beats[1]
        v.get_children()[1]._filled_quarter_duration = 1 / 4
        assert v.get_current_beat() == beats[2]
        v.get_children()[2]._filled_quarter_duration = 1 / 4
        assert v.get_current_beat() == beats[3]
        v.get_children()[3]._filled_quarter_duration = 1 / 8
        assert v.get_current_beat() == beats[3]
        v.get_children()[3]._filled_quarter_duration = 1 / 4
        with self.assertRaises(VoiceIsFullError):
            assert v.get_current_beat() is None

    @patch("musicscore.staff.Staff")
    def test_add_chord(self, mock_staff):
        v = Voice()
        v._parent = mock_staff

        with self.assertRaises(VoiceHasNoBeatsError):
            v._add_chord(Chord(60, 1))
        v.update_beats(1, 1, 1, 1)
        v._add_chord(Chord(quarter_duration=1.5, midis=60))
        v._add_chord(Chord(quarter_duration=2, midis=60))
        v._add_chord(Chord(quarter_duration=0.5, midis=60))
        assert v.leftover_chord is None
        assert [ch.quarter_duration for ch in v.get_chords()] == [1.5, 0.5, 1.5, 0.5]
        assert v.get_chords()[1].midis[0].is_tied_to_next
        assert v.get_chords()[2].midis[0].is_tied_to_previous

        v = Voice()
        v._parent = mock_staff
        v.update_beats(1, 1)
        v._add_chord(Chord(quarter_duration=3, midis=60))
        assert isinstance(v.leftover_chord, Chord)
        assert v.leftover_chord.quarter_duration == 1

    def test_add_chord_exception(self):
        v = Voice()
        with self.assertRaises(AddChordError):
            v.add_chord()



"""
Different kinds of grace notes: acciaccatura, appoggiatura; beamed grace notes; grace notes with
accidentals; different durations of the grace notes.
"""

from pathlib import Path

from musicscore import Score, C, Chord, F, E, D, A
from musicscore.chord import GraceChord
from unittest import TestCase
from musicxml.xmlelement.xmlelement import XMLSlur


class TestLily24a(TestCase):
    def test_lily_24a_GraceNote(self):
        score = Score()
        p = score.add_part("p1")
        measure_1_chords = [Chord(C(5), qd) for qd in [1, 1, 1, 1]]

        measure_1_chords[0].add_grace_chord(D(5), "16th")

        measure_1_chords[1].add_grace_chord(E(5), "16th")
        measure_1_chords[1].add_grace_chord(D(5), "16th")

        measure_1_chords[2].add_grace_chord(D(5), "16th")

        measure_1_chords[3].add_grace_chord(D(5), "eighth")

        measure_2_chords = [Chord(C(5), qd) for qd in [1, 2, 0.5, 0.5]]

        measure_2_chords[0].add_grace_chord(D(5), "16th").add_x(XMLSlur(type="start"))
        measure_2_chords[0].add_x(XMLSlur(type="stop"))

        measure_2_chords[1].add_grace_chord(E(5), "16th")
        measure_2_chords[1].add_grace_chord(D(5), "16th")

        measure_2_chords[2].add_grace_chord(D(5), "16th").add_x(XMLSlur(type="start"))
        measure_2_chords[2].add_x(XMLSlur(type="stop"))

        measure_2_chords[3].add_grace_chord(D(5), "16th").add_x(XMLSlur(type="start"))
        measure_2_chords[3].add_x(XMLSlur(type="stop"))
        measure_2_chords[3].add_grace_chord(E(5), "16th", position="after")

        measure_3_chords = [
            Chord([F(4), C(5)], 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
        ]

        measure_3_chords[0].add_grace_chord(E(5), "16th")

        measure_3_chords[1].add_grace_chord(GraceChord(D(5, "#"), type="quarter"))

        measure_3_chords[2].add_grace_chord(GraceChord(D(5, "b"), type="quarter"))
        measure_3_chords[2].add_grace_chord(GraceChord(A(4, "b"), type="quarter"))

        for chord in measure_1_chords + measure_2_chords + measure_3_chords:
            p.add_chord(chord)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from musicscore import Chord, QuarterDuration, Part
from unittest import TestCase
from musicscore.tests.util_subdivisions import generate_all_subdivision_patterns


class TestBeamsAndPrintedDurations(TestCase):
    def test_quarter_32th(self):
        p = Part("p1")
        p.add_measure([1, 4])
        subdivision = 8
        rhythmic_patterns = generate_all_subdivision_patterns(subdivision, True)
        for pattern in rhythmic_patterns:
            [p.add_chord(Chord(60, QuarterDuration(x, subdivision))) for x in pattern]
        p.finalize()
        for beat in p.get_beats():
            beat_chords = beat.get_chords()
            for i in range(1, len(beat_chords)):
                current_chord = beat_chords[i]
                previous_chord = beat_chords[i - 1]
                current_chord.check_printed_duration()
                current_chord.check_number_of_beams()
                if current_chord.offset == 1 / 2:
                    for key, value in current_chord.beams.items():
                        if key == 1:
                            assert value in ["end", "continue"]
                        else:
                            assert value != "continue"
                    for key, value in previous_chord.beams.items():
                        if key == 1:
                            assert value in ["begin", "continue"]
                        else:
                            assert value != "continue"
                    continue

    def test_quarter_nonuplets(self):
        p = Part("p1")
        p.add_measure([1, 4])
        subdivision = 9
        rhythmic_patterns = generate_all_subdivision_patterns(subdivision, True)
        for pattern in rhythmic_patterns:
            [p.add_chord(Chord(60, QuarterDuration(x, subdivision))) for x in pattern]
        p.finalize()
        for beat in p.get_beats():
            beat_chords = beat.get_chords()
            for chord in beat_chords:
                chord.check_printed_duration()
                chord.check_number_of_beams()



from pathlib import Path

from fractions import Fraction

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from unittest import TestCase
from musicscore.tests.util_subdivisions import (
    generate_all_16ths_manually,
    generate_all_32nds_manually,
)


class TestBeams1(TestCase):
    def test_beams_1(self):
        """
        Write all possible combinations of 16ths and 32nds in a beat
        """
        s = Score()
        p = s.add_child(Part("P1", name="Music"))

        p.add_measure(time=(1, 4))

        groups = (
            [(Fraction(1, 2), Fraction(1, 2))]
            + generate_all_16ths_manually()
            + generate_all_32nds_manually()
        )
        for group in groups:
            for qd in group:
                p.add_chord(Chord(60, qd))
        xml_path = Path(__file__).with_suffix(".xml")
        s.export_xml(xml_path)



from unittest import TestCase

from musicscore.exceptions import TupletNormalTypeError
from musicscore.tuplet import Tuplet
from musicxml import XMLTimeModification


class TestTuplet(TestCase):
    def test_init(self):
        t = Tuplet(
            actual_notes=3,
            normal_notes=2,
            normal_type="eighth",
            bracket_type="start",
            bracket_number=1,
        )
        assert t.actual_notes == 3
        assert t.normal_notes == 2
        assert t.normal_type == "eighth"
        assert t.bracket_type == "start"
        assert t.bracket_number == 1

    def test_property_exceptions(self):
        t = Tuplet()
        with self.assertRaises(ValueError):
            t.actual_notes = -1
        with self.assertRaises(ValueError):
            t.actual_notes = 0
        with self.assertRaises(ValueError):
            t.actual_notes = 1
        with self.assertRaises(TypeError):
            t.actual_notes = None
        with self.assertRaises(TypeError):
            t.actual_notes = "1"

        with self.assertRaises(ValueError):
            t.normal_notes = -1
        with self.assertRaises(ValueError):
            t.normal_notes = 0
        with self.assertRaises(ValueError):
            t.normal_notes = 1

        with self.assertRaises(TypeError):
            t.normal_notes = "1"

        with self.assertRaises(ValueError):
            t.normal_type = "bla"

        with self.assertRaises(ValueError):
            t.bracket_type = "bla"

        with self.assertRaises(ValueError):
            t.bracket_number = -1
        with self.assertRaises(ValueError):
            t.bracket_number = 0
        with self.assertRaises(TypeError):
            t.bracket_number = "1"

    def test_default_values(self):
        t = Tuplet()
        assert t.actual_notes == 3
        assert t.normal_notes == 2
        assert t.normal_type == "eighth"
        assert t.bracket_type is None
        assert t.bracket_number is None

    def test_tuplet_get_xml_time_modification(self):
        t = Tuplet()
        tm = t.get_xml_time_modification()
        assert isinstance(tm, XMLTimeModification)
        assert tm.xml_actual_notes.value_ == t.actual_notes
        assert tm.xml_normal_notes.value_ == t.normal_notes
        assert tm.xml_normal_type.value_ == t.normal_type

    def test_tuplet_get_xml_tuplet(self):
        t = Tuplet()
        assert t.get_xml_tuplet() is None
        assert t.bracket_number is None
        t.bracket_type = "start"
        assert t.bracket_number == 1
        assert t.get_xml_tuplet().bracket == "yes"
        assert t.get_xml_tuplet().type == t.bracket_type
        assert t.get_xml_tuplet().number == 1
        t.bracket_number = 2
        assert t.bracket_number == 2
        assert t.get_xml_tuplet().number == 2

    def test_tuplet_quarter_duration(self):
        t = Tuplet()
        assert t.quarter_duration == 1
        t.quarter_duration = 1.2
        with self.assertRaises(TupletNormalTypeError):
            t.normal_type

    def test_tuplet_automatic_type(self):
        t = Tuplet(quarter_duration=2)
        assert t.normal_type == "quarter"
        t.quarter_duration = 1 / 2
        assert t.normal_type == "16th"
